# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import abc

import six

from rally.common import broker
from rally.common.i18n import _
from rally.common import logging
from rally.common import utils
from rally import consts
from rally import osclients
from rally.plugins.openstack.scenarios.nova import utils as nova_utils
from rally.plugins.openstack.scenarios.vm import vmtasks
from rally.task import context
from rally.task import types

LOG = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
@context.configure(name="custom_image", order=500, hidden=True)
class BaseCustomImageGenerator(context.Context):
    """Base class for the contexts providing customized image with.

    Every context class for the specific customization must implement
    the method `_customize_image` that is able to connect to the server
    using SSH and e.g. install applications inside it.

    This is used e.g. to install the benchmark application using SSH
    access.

    This base context class provides a way to prepare an image with
    custom preinstalled applications. Basically, this code boots a VM, calls
    the `_customize_image` and then snapshots the VM disk, removing the VM
    afterwards. The image UUID is stored in the user["custom_image"]["id"]
    and can be used afterwards by scenario.
    """

    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA,
        "properties": {
            "image": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string"
                    }
                }
            },
            "flavor": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string"
                    }
                }
            },
            "username": {
                "type": "string"
            },
            "password": {
                "type": "string"
            },
            "floating_network": {
                "type": "string"
            },
            "internal_network": {
                "type": "string"
            },
            "port": {
                "type": "integer",
                "minimum": 1,
                "maximum": 65535
            },
            "userdata": {
                "type": "string"
            },
            "workers": {
                "type": "integer",
                "minimum": 1,
            }
        },
        "required": ["image", "flavor"],
        "additionalProperties": False
    }

    DEFAULT_CONFIG = {
        "username": "root",
        "port": 22,
        "workers": 1
    }

    @logging.log_task_wrapper(LOG.info, _("Enter context: `custom_image`"))
    def setup(self):
        """Creates custom image(s) with preinstalled applications.

        When admin is present creates one public image that is usable
        from all the tenants and users. Otherwise create one image
        per user and tenant.
        """

        if "admin" in self.context:
            # NOTE(pboldin): Create by first user and make it public by
            #                the admin
            user = self.context["users"][0]
            tenant = self.context["tenants"][user["tenant_id"]]

            nics = None
            if "networks" in tenant:
                nics = [{"net-id": tenant["networks"][0]["id"]}]

            custom_image = self.create_one_image(user, nics=nics)
            self.make_image_public(custom_image)

            for tenant in self.context["tenants"].values():
                tenant["custom_image"] = custom_image
        else:
            def publish(queue):
                users = self.context.get("users", [])
                for user, tenant_id in utils.iterate_per_tenants(users):
                    queue.append((user, tenant_id))

            def consume(cache, args):
                user, tenant_id = args
                tenant = self.context["tenants"][tenant_id]
                tenant["custom_image"] = self.create_one_image(user)

            broker.run(publish, consume, self.config["workers"])

    def create_one_image(self, user, **kwargs):
        """Create one image for the user."""

        clients = osclients.Clients(user["credential"])

        image_id = types.ImageResourceType.transform(
            clients=clients, resource_config=self.config["image"])
        flavor_id = types.FlavorResourceType.transform(
            clients=clients, resource_config=self.config["flavor"])

        vm_scenario = vmtasks.VMTasks(self.context, clients=clients)

        server, fip = vm_scenario._boot_server_with_fip(
            image=image_id, flavor=flavor_id,
            floating_network=self.config.get("floating_network"),
            userdata=self.config.get("userdata"),
            key_name=user["keypair"]["name"],
            security_groups=[user["secgroup"]["name"]],
            **kwargs)

        try:
            LOG.debug("Installing benchmark on %r %s", server, fip["ip"])
            self.customize_image(server, fip, user)

            LOG.debug("Stopping server %r", server)
            vm_scenario._stop_server(server)

            LOG.debug("Creating snapshot for %r", server)
            custom_image = vm_scenario._create_image(server).to_dict()
        finally:
            vm_scenario._delete_server_with_fip(server, fip)

        return custom_image

    def make_image_public(self, custom_image):
        """Make the image available publicly."""

        admin_clients = osclients.Clients(self.context["admin"]["credential"])

        LOG.debug("Making image %r public", custom_image["id"])
        admin_clients.glance().images.get(
            custom_image["id"]).update(is_public=True)

    @logging.log_task_wrapper(LOG.info, _("Exit context: `custom_image`"))
    def cleanup(self):
        """Delete created custom image(s)."""

        if "admin" in self.context:
            user = self.context["users"][0]
            tenant = self.context["tenants"][user["tenant_id"]]
            if "custom_image" in tenant:
                self.delete_one_image(user, tenant["custom_image"])
                tenant.pop("custom_image")
        else:
            def publish(queue):
                users = self.context.get("users", [])
                for user, tenant_id in utils.iterate_per_tenants(users):
                    queue.append((user, tenant_id))

            def consume(cache, args):
                user, tenant_id = args
                tenant = self.context["tenants"][tenant_id]
                if "custom_image" in tenant:
                    self.delete_one_image(user, tenant["custom_image"])
                    tenant.pop("custom_image")

            broker.run(publish, consume, self.config["workers"])

    def delete_one_image(self, user, custom_image):
        """Delete the image created for the user and tenant."""

        clients = osclients.Clients(user["credential"])

        nova_scenario = nova_utils.NovaScenario(
            context=self.context, clients=clients)

        with logging.ExceptionLogger(
                LOG, _("Unable to delete image %s") % custom_image["id"]):

            custom_image = nova_scenario.clients("nova").images.get(
                custom_image["id"])
            nova_scenario._delete_image(custom_image)

    @logging.log_task_wrapper(LOG.info,
                              _("Custom image context: customizing"))
    def customize_image(self, server, ip, user):
        return self._customize_image(server, ip, user)

    @abc.abstractmethod
    def _customize_image(self, server, ip, user):
        """Override this method with one that customizes image.

        Basically, code can simply call `VMScenario._run_command` function
        specifying an installation script and interpreter. This script will
        be then executed using SSH.

        :param server: nova.Server instance
        :param ip: dict with server IP details
        :param user: user who started a VM instance. Used to extract keypair
        """
        pass



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally.plugins.openstack.scenarios.murano import environments
from tests.unit import test

CTX = "rally.task.context"
MURANO_SCENARIO = ("rally.plugins.openstack.scenarios.murano."
                   "environments.MuranoEnvironments")


class MuranoEnvironmentsTestCase(test.ScenarioTestCase):

    def _get_context(self):
        self.context.update({
            "tenant": {
                "packages": [mock.MagicMock(fully_qualified_name="fake")]
            },
            "user": {
                "tenant_id": "fake_tenant_id"
            },
            "config": {
                "murano_packages": {
                    "app_package": (
                        "rally-jobs/extra/murano/"
                        "applications/HelloReporter/"
                        "io.murano.apps.HelloReporter.zip")
                }
            }
        })
        return self.context

    @mock.patch(MURANO_SCENARIO + "._list_environments")
    def test_list_environments(self, mock__list_environments):
        scenario = environments.MuranoEnvironments(self.context)
        scenario._list_environments()
        mock__list_environments.assert_called_once_with()

    @mock.patch(MURANO_SCENARIO + "._create_session")
    @mock.patch(MURANO_SCENARIO + "._delete_environment")
    @mock.patch(MURANO_SCENARIO + "._create_environment")
    @mock.patch(MURANO_SCENARIO + ".generate_random_name")
    def test_create_and_delete_environment(
            self, mock_generate_random_name, mock__create_environment,
            mock__delete_environment, mock__create_session):
        scenario = environments.MuranoEnvironments(self.context)
        fake_environment = mock.Mock(id="fake_id")
        mock__create_environment.return_value = fake_environment
        mock_generate_random_name.return_value = "foo"
        scenario.create_and_delete_environment()
        mock__create_environment.assert_called_once_with()
        mock__create_session.assert_called_once_with(fake_environment.id)
        mock__delete_environment.assert_called_once_with(fake_environment)

    @mock.patch(MURANO_SCENARIO + "._create_environment")
    @mock.patch(MURANO_SCENARIO + "._create_session")
    @mock.patch(MURANO_SCENARIO + "._create_service")
    @mock.patch(MURANO_SCENARIO + "._deploy_environment")
    def test_create_and_deploy_environment(
            self, mock__deploy_environment, mock__create_service,
            mock__create_session, mock__create_environment):

        fake_environment = mock.MagicMock(id="fake_env_id")
        mock__create_environment.return_value = fake_environment

        fake_session = mock.Mock(id="fake_session_id")
        mock__create_session.return_value = fake_session

        scenario = environments.MuranoEnvironments(self.context)
        scenario.context = self._get_context()
        scenario.context["tenants"] = {
            "fake_tenant_id": {
                "packages": [mock.MagicMock()]
            }
        }

        scenario.create_and_deploy_environment(1)

        mock__create_environment.assert_called_once_with()
        mock__create_session.assert_called_once_with(fake_environment.id)
        mock__create_service.assert_called_once_with(
            fake_environment, fake_session, "fake", atomic_action=False)
        mock__deploy_environment.assert_called_once_with(
            fake_environment, fake_session)



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function

from rally.cli import cliutils
from rally.cli.commands import plugin


class InfoCommands(object):
    """[Deprecated since 0.1.1] Allows you to get quick doc of rally entities.

    """

    @cliutils.args("--query", dest="query", type=str, help="Search query.")
    def find(self, query):
        """Search for an entity that matches the query and print info about it.

        :param query: search query.
        """
        print("This command was deprecated, and will be removed in 0.2.0 use:")
        print("rally plugin show %s" % query)

        plugin.PluginCommands().show(query)
        return 1

    def list(self):
        """List main entities in Rally for which rally info find works.

        Lists task scenario groups, deploy engines and server providers.
        """
        print("This command was deprecated, and will be removed in 0.2.0 use:")
        print("rally plugin list")

        plugin.PluginCommands().list()
        return 1



# Copyright 2015: Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tempfile

from rally import consts
from rally.plugins.openstack import scenario
from rally.plugins.openstack.scenarios.swift import utils
from rally.task import atomic
from rally.task import validation


class SwiftObjects(utils.SwiftScenario):
    """Benchmark scenarios for Swift Objects."""

    @validation.required_services(consts.Service.SWIFT)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"cleanup": ["swift"]})
    def create_container_and_object_then_list_objects(
            self, objects_per_container=1,
            object_size=1024, **kwargs):
        """Create container and objects then list all objects.

        :param objects_per_container: int, number of objects to upload
        :param object_size: int, temporary local object size
        :param kwargs: dict, optional parameters to create container
        """
        key_suffix = "object"
        if objects_per_container > 1:
            key_suffix = "%i_objects" % objects_per_container

        container_name = None
        with tempfile.TemporaryFile() as dummy_file:
            # set dummy file to specified object size
            dummy_file.truncate(object_size)
            container_name = self._create_container(**kwargs)
            with atomic.ActionTimer(self, "swift.create_%s" % key_suffix):
                for i in range(objects_per_container):
                    dummy_file.seek(0)
                    self._upload_object(container_name, dummy_file,
                                        atomic_action=False)
        self._list_objects(container_name)

    @validation.required_services(consts.Service.SWIFT)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"cleanup": ["swift"]})
    def create_container_and_object_then_delete_all(
            self, objects_per_container=1,
            object_size=1024, **kwargs):
        """Create container and objects then delete everything created.

        :param objects_per_container: int, number of objects to upload
        :param object_size: int, temporary local object size
        :param kwargs: dict, optional parameters to create container
        """
        key_suffix = "object"
        if objects_per_container > 1:
            key_suffix = "%i_objects" % objects_per_container

        container_name = None
        objects_list = []
        with tempfile.TemporaryFile() as dummy_file:
            # set dummy file to specified object size
            dummy_file.truncate(object_size)
            container_name = self._create_container(**kwargs)
            with atomic.ActionTimer(self, "swift.create_%s" % key_suffix):
                for i in range(objects_per_container):
                    dummy_file.seek(0)
                    object_name = self._upload_object(container_name,
                                                      dummy_file,
                                                      atomic_action=False)[1]
                    objects_list.append(object_name)

        with atomic.ActionTimer(self, "swift.delete_%s" % key_suffix):
            for object_name in objects_list:
                self._delete_object(container_name, object_name,
                                    atomic_action=False)
        self._delete_container(container_name)

    @validation.required_services(consts.Service.SWIFT)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"cleanup": ["swift"]})
    def create_container_and_object_then_download_object(
            self, objects_per_container=1,
            object_size=1024, **kwargs):
        """Create container and objects then download all objects.

        :param objects_per_container: int, number of objects to upload
        :param object_size: int, temporary local object size
        :param kwargs: dict, optional parameters to create container
        """
        key_suffix = "object"
        if objects_per_container > 1:
            key_suffix = "%i_objects" % objects_per_container

        container_name = None
        objects_list = []
        with tempfile.TemporaryFile() as dummy_file:
            # set dummy file to specified object size
            dummy_file.truncate(object_size)
            container_name = self._create_container(**kwargs)
            with atomic.ActionTimer(self, "swift.create_%s" % key_suffix):
                for i in range(objects_per_container):
                    dummy_file.seek(0)
                    object_name = self._upload_object(container_name,
                                                      dummy_file,
                                                      atomic_action=False)[1]
                    objects_list.append(object_name)

        with atomic.ActionTimer(self, "swift.download_%s" % key_suffix):
            for object_name in objects_list:
                self._download_object(container_name, object_name,
                                      atomic_action=False)

    @validation.required_services(consts.Service.SWIFT)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"swift_objects": {}})
    def list_objects_in_containers(self):
        """List objects in all containers."""
        containers = self._list_containers()[1]

        key_suffix = "container"
        if len(containers) > 1:
            key_suffix = "%i_containers" % len(containers)

        with atomic.ActionTimer(self, "swift.list_objects_in_%s" % key_suffix):
            for container in containers:
                self._list_objects(container["name"], atomic_action=False)

    @validation.required_services(consts.Service.SWIFT)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"swift_objects": {}})
    def list_and_download_objects_in_containers(self):
        """List and download objects in all containers."""
        containers = self._list_containers()[1]

        list_key_suffix = "container"
        if len(containers) > 1:
            list_key_suffix = "%i_containers" % len(containers)

        objects_dict = {}
        with atomic.ActionTimer(self,
                                "swift.list_objects_in_%s" % list_key_suffix):
            for container in containers:
                container_name = container["name"]
                objects_dict[container_name] = self._list_objects(
                    container_name,
                    atomic_action=False)[1]

        objects_total = sum(map(len, objects_dict.values()))
        download_key_suffix = "object"
        if objects_total > 1:
            download_key_suffix = "%i_objects" % objects_total

        with atomic.ActionTimer(self,
                                "swift.download_%s" % download_key_suffix):
            for container_name, objects in objects_dict.items():
                for obj in objects:
                    self._download_object(container_name, obj["name"],
                                          atomic_action=False)



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import multiprocessing
import time

import jsonschema

from rally.common import logging
from rally.common import objects
from rally.common.plugin import plugin
from rally.common import utils as rutils
from rally import consts
from rally.task import context
from rally.task import scenario
from rally.task import types
from rally.task import utils


LOG = logging.getLogger(__name__)


def format_result_on_timeout(exc, timeout):
    return {
        "duration": timeout,
        "idle_duration": 0,
        "output": {"additive": [], "complete": []},
        "atomic_actions": {},
        "error": utils.format_exc(exc)
    }


def _get_scenario_context(context_obj):
    return context.ContextManager(context_obj).map_for_scenario()


def _run_scenario_once(args):
    iteration, cls, method_name, context_obj, kwargs = args

    LOG.info("Task %(task)s | ITER: %(iteration)s START" %
             {"task": context_obj["task"]["uuid"], "iteration": iteration})

    context_obj["iteration"] = iteration
    scenario_inst = cls(context_obj)

    error = []
    output = {"additive": [], "complete": []}
    try:
        with rutils.Timer() as timer:
            # NOTE(amaretskiy): Output as return value is deprecated
            #     but supported for backward compatibility
            deprecated_output = getattr(scenario_inst, method_name)(**kwargs)
            warning = ""
            if deprecated_output:
                warning = ("Returning output data by scenario is deprecated "
                           "in favor of calling add_output().")
            if scenario_inst._output != {"complete": [], "additive": []}:
                output = scenario_inst._output
                if deprecated_output:
                    warning += (" Output data both returned and passed to "
                                "add_output() so returned one is ignored!")
            elif deprecated_output:
                output["additive"].append({
                    "title": "Scenario output",
                    "description": "",
                    "chart_plugin": "StackedArea",
                    "data": [list(item)
                             for item in deprecated_output["data"].items()]})
            if warning:
                LOG.warning(warning)
    except Exception as e:
        error = utils.format_exc(e)
        if logging.is_debug():
            LOG.exception(e)
    finally:
        status = "Error %s: %s" % tuple(error[0:2]) if error else "OK"
        LOG.info("Task %(task)s | ITER: %(iteration)s END: %(status)s" %
                 {"task": context_obj["task"]["uuid"], "iteration": iteration,
                  "status": status})

        return {"duration": timer.duration() - scenario_inst.idle_duration(),
                "timestamp": timer.timestamp(),
                "idle_duration": scenario_inst.idle_duration(),
                "error": error,
                "output": output,
                "atomic_actions": scenario_inst.atomic_actions()}


def _worker_thread(queue, args):
    queue.put(_run_scenario_once(args))


def _log_worker_info(**info):
    """Log worker parameters for debugging.

    :param info: key-value pairs to be logged
    """
    info_message = "\n\t".join(["%s: %s" % (k, v)
                                for k, v in info.items()])
    LOG.debug("Starting a worker.\n\t%s" % info_message)


class ScenarioRunnerResult(dict):
    """Class for all scenario runners' result."""

    RESULT_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA,
        "properties": {
            "duration": {
                "type": "number"
            },
            "timestamp": {
                "type": "number"
            },
            "idle_duration": {
                "type": "number"
            },
            "output": objects.task.OUTPUT_SCHEMA,
            "atomic_actions": {
                "type": "object",
                "patternProperties": {
                    ".*": {"type": ["number", "null"]}
                }
            },
            "error": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            }
        },
        "additionalProperties": False
    }

    def __init__(self, result_list):
        super(ScenarioRunnerResult, self).__init__(result_list)
        jsonschema.validate(result_list, self.RESULT_SCHEMA)


def configure(name, namespace="default"):
    return plugin.configure(name=name, namespace=namespace)


@configure(name="base_runner")
class ScenarioRunner(plugin.Plugin):
    """Base class for all scenario runners.

    Scenario runner is an entity that implements a certain strategy of
    launching benchmark scenarios, e.g. running them continuously or
    periodically for a given number of times or seconds.
    These strategies should be implemented in subclasses of ScenarioRunner
    in the_run_scenario() method.
    """

    CONFIG_SCHEMA = {}

    def __init__(self, task, config, batch_size=0):
        """Runner constructor.

        It sets task and config to local variables. Also initialize
        result_queue, where results will be put by _send_result method.

        :param task: Instance of objects.Task
        :param config: Dict with runner section from benchmark configuration
        """
        self.task = task
        self.config = config
        self.result_queue = collections.deque()
        self.aborted = multiprocessing.Event()
        self.run_duration = 0
        self.batch_size = batch_size
        self.result_batch = []

    @staticmethod
    def validate(config):
        """Validates runner's part of task config."""
        runner = ScenarioRunner.get(config.get("type", "serial"))
        jsonschema.validate(config, runner.CONFIG_SCHEMA)

    @abc.abstractmethod
    def _run_scenario(self, cls, method_name, context, args):
        """Runs the specified benchmark scenario with given arguments.

        :param cls: The Scenario class where the scenario is implemented
        :param method_name: Name of the method that implements the scenario
        :param context: Benchmark context that contains users, admin & other
                        information, that was created before benchmark started.
        :param args: Arguments to call the scenario method with

        :returns: List of results fore each single scenario iteration,
                  where each result is a dictionary
        """

    def run(self, name, context, args):
        cls_name, method_name = name.split(".", 1)
        cls = scenario.Scenario.get(name)._meta_get("cls_ref")

        # NOTE(boris-42): processing @types decorators
        args = types.preprocess(name, context, args)

        with rutils.Timer() as timer:
            self._run_scenario(cls, method_name, context, args)

        self.run_duration = timer.duration()
        return self.run_duration

    def abort(self):
        """Abort the execution of further benchmark scenario iterations."""
        self.aborted.set()

    @staticmethod
    def _create_process_pool(processes_to_start, worker_process,
                             worker_args_gen):
        """Create a pool of processes with some defined target function.

        :param processes_to_start: number of processes to create in the pool
        :param worker_process: target function for all processes in the pool
        :param worker_args_gen: generator of arguments for the target function
        :returns: the process pool as a deque
        """
        process_pool = collections.deque()

        for i in range(processes_to_start):
            kwrgs = {"processes_to_start": processes_to_start,
                     "processes_counter": i}
            process = multiprocessing.Process(target=worker_process,
                                              args=next(worker_args_gen),
                                              kwargs={"info": kwrgs})
            process.start()
            process_pool.append(process)

        return process_pool

    def _join_processes(self, process_pool, result_queue):
        """Join the processes in the pool and send their results to the queue.

        :param process_pool: pool of processes to join
        :result_queue: multiprocessing.Queue that receives the results
        """
        while process_pool:
            while process_pool and not process_pool[0].is_alive():
                process_pool.popleft().join()

            if result_queue.empty():
                # sleep a bit to avoid 100% usage of CPU by this method
                time.sleep(0.001)

            while not result_queue.empty():
                self._send_result(result_queue.get())

        self._flush_results()
        result_queue.close()

    def _flush_results(self):
        if self.result_batch:
            sorted_batch = sorted(self.result_batch)
            self.result_queue.append(sorted_batch)
            self.result_batch = []

    def _send_result(self, result):
        """Store partial result to send it to consumer later.

        :param result: Result dict to be sent. It should match the
                       ScenarioRunnerResult schema, otherwise
                       ValidationError is raised.
        """

        r = ScenarioRunnerResult(result)
        self.result_batch.append(r)

        if len(self.result_batch) >= self.batch_size:
            sorted_batch = sorted(self.result_batch,
                                  key=lambda r: r["timestamp"])
            self.result_queue.append(sorted_batch)
            self.result_batch = []

    def _log_debug_info(self, **info):
        """Log runner parameters for debugging.

        The method logs the runner name, the task id as well as the values
        passed as arguments.

        :param info: key-value pairs to be logged
        """
        info_message = "\n\t".join(["%s: %s" % (k, v)
                                    for k, v in info.items()])
        LOG.debug("Starting the %(name)s runner (task UUID: %(task)s)."
                  "\n\t%(info)s" %
                  {"name": self._meta_get("name"),
                   "task": self.task["uuid"],
                   "info": info_message})



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally.plugins.openstack.scenarios.ceilometer import stats
from tests.unit import test


class CeilometerStatsTestCase(test.ScenarioTestCase):

    def test_get_stats(self):
        scenario = stats.CeilometerStats(self.context)
        scenario._get_stats = mock.MagicMock()
        context = {"user": {"tenant_id": "fake", "id": "fake_id"},
                   "tenant": {"id": "fake_id",
                              "resources": ["fake_resource"]}}
        metadata_query = {"a": "test"}
        period = 10
        groupby = "user_id"
        aggregates = "sum"
        scenario.context = context
        scenario.get_stats("fake_meter", True, True, True, metadata_query,
                           period, groupby, aggregates)
        scenario._get_stats.assert_called_once_with(
            "fake_meter",
            [{"field": "user_id", "value": "fake_id", "op": "eq"},
             {"field": "project_id", "value": "fake_id", "op": "eq"},
             {"field": "resource_id", "value": "fake_resource", "op": "eq"},
             {"field": "metadata.a", "value": "test", "op": "eq"}],
            10,
            "user_id",
            "sum"
        )



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import uuid

import mock
from oslo_config import fixture
from oslotest import base
from oslotest import mockpatch

from rally.common import db
from rally import plugins
from tests.unit import fakes


class DatabaseFixture(fixture.Config):
    """Create clean DB before starting test."""
    def setUp(self):
        super(DatabaseFixture, self).setUp()
        db_url = os.environ.get("RALLY_UNITTEST_DB_URL", "sqlite://")
        db.db_cleanup()
        self.conf.set_default("connection", db_url, group="database")
        db.db_drop()
        db.db_create()


class TestCase(base.BaseTestCase):
    """Test case base class for all unit tests."""

    def setUp(self):
        super(TestCase, self).setUp()
        self.addCleanup(mock.patch.stopall)
        plugins.load()

    def _test_atomic_action_timer(self, atomic_actions, name):
        action_duration = atomic_actions.get(name)
        self.assertIsNotNone(action_duration)
        self.assertIsInstance(action_duration, float)

    def assertSequenceEqual(self, iterable_1, iterable_2, msg=None):
        self.assertEqual(tuple(iterable_1), tuple(iterable_2), msg)


class DBTestCase(TestCase):
    """Base class for tests which use DB."""

    def setUp(self):
        super(DBTestCase, self).setUp()
        self.useFixture(DatabaseFixture())


# TODO(boris-42): This should be moved to test.plugins.test module
#                 or similar

class ScenarioTestCase(TestCase):
    """Base class for Scenario tests using mocked self.clients."""
    benchmark_utils = "rally.task.utils"
    patch_benchmark_utils = True

    def client_factory(self, client_type, version=None, admin=False):
        """Create a new client object."""
        return mock.MagicMock(client_type=client_type,
                              version=version,
                              admin=admin)

    def clients(self, client_type, version=None, admin=False):
        """Get a mocked client."""
        key = (client_type, version, admin)
        if key not in self._clients:
            self._clients[key] = self.client_factory(client_type,
                                                     version=version,
                                                     admin=admin)
        return self._clients[key]

    def admin_clients(self, client_type, version=None):
        """Get a mocked admin client."""
        return self.clients(client_type, version=version, admin=True)

    def client_created(self, client_type, version=None, admin=False):
        """Determine if a client has been created.

        This can be used to see if a scenario calls
        'self.clients("foo")', without checking to see what was done
        with the client object returned by that call.
        """
        key = (client_type, version, admin)
        return key in self._clients

    def setUp(self):
        super(ScenarioTestCase, self).setUp()
        if self.patch_benchmark_utils:
            self.mock_resource_is = mockpatch.Patch(
                self.benchmark_utils + ".resource_is")
            self.mock_get_from_manager = mockpatch.Patch(
                self.benchmark_utils + ".get_from_manager")
            self.mock_wait_for = mockpatch.Patch(
                self.benchmark_utils + ".wait_for")
            self.mock_wait_for_delete = mockpatch.Patch(
                self.benchmark_utils + ".wait_for_delete")
            self.mock_wait_for_status = mockpatch.Patch(
                self.benchmark_utils + ".wait_for_status")
            self.useFixture(self.mock_resource_is)
            self.useFixture(self.mock_get_from_manager)
            self.useFixture(self.mock_wait_for)
            self.useFixture(self.mock_wait_for_delete)
            self.useFixture(self.mock_wait_for_status)

        self.mock_sleep = mockpatch.Patch("time.sleep")
        self.useFixture(self.mock_sleep)

        self._clients = {}
        base_path = "rally.plugins.openstack"
        self._client_mocks = [
            mock.patch(
                "%s.scenario.OpenStackScenario.clients" % base_path,
                mock.Mock(side_effect=self.clients)),
            mock.patch(
                "%s.scenario.OpenStackScenario.admin_clients" % base_path,
                mock.Mock(side_effect=self.admin_clients))
        ]
        for patcher in self._client_mocks:
            patcher.start()

        self.context = get_test_context()

    def tearDown(self):
        for patcher in self._client_mocks:
            patcher.stop()
        super(ScenarioTestCase, self).tearDown()


class FakeClientsScenarioTestCase(ScenarioTestCase):
    """Base class for Scenario tests using fake (not mocked) self.clients."""

    def client_factory(self, client_type, version=None, admin=False):
        return getattr(self._fake_clients, client_type)()

    def setUp(self):
        super(FakeClientsScenarioTestCase, self).setUp()
        self._fake_clients = fakes.FakeClients()


def get_test_context():
    return {
        "task": {
            "uuid": str(uuid.uuid4())
        }
    }



# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from rally.plugins.openstack.context.sahara import sahara_output_data_sources
from tests.unit import test

CTX = "rally.plugins.openstack.context.sahara"


class FakeDict(object):
    def __init__(self):
        pass


class SaharaOutputDataSourcesTestCase(test.ScenarioTestCase):

    def setUp(self):
        super(SaharaOutputDataSourcesTestCase, self).setUp()
        fake_dict = FakeDict()
        fake_dict.__dict__["username"] = "user"
        fake_dict.__dict__["password"] = "passwd"
        self.tenants_num = 2
        self.users_per_tenant = 2
        self.users = self.tenants_num * self.users_per_tenant
        self.task = mock.MagicMock()

        self.tenants = {}
        self.users_key = []

        for i in range(self.tenants_num):
            self.tenants[str(i)] = {"id": str(i), "name": str(i),
                                    "sahara": {"image": "42"}}
            for j in range(self.users_per_tenant):
                self.users_key.append({"id": "%s_%s" % (str(i), str(j)),
                                       "tenant_id": str(i),
                                       "credential": fake_dict})

        self.user_key = [{"id": i, "tenant_id": j, "credential": "credential"}
                         for j in range(self.tenants_num)
                         for i in range(self.users_per_tenant)]
        self.context.update({
            "config": {
                "users": {
                    "tenants": self.tenants_num,
                    "users_per_tenant": self.users_per_tenant,
                },
                "sahara_output_data_sources": {
                    "output_type": "hdfs",
                    "output_url_prefix": "hdfs://test_host/",
                },
            },
            "admin": {"credential": mock.MagicMock()},
            "task": mock.MagicMock(),
            "users": self.users_key,
            "tenants": self.tenants,
        })

    def check_setup(self):
        context = sahara_output_data_sources.SaharaOutputDataSources.context[
            "sahara"]["output_conf"]
        self.assertIsNotNone(context.get("output_type"))
        self.assertIsNotNone(context.get("output_url_prefix"))

    @mock.patch("%s.sahara_output_data_sources.resource_manager.cleanup" % CTX)
    @mock.patch("%s.sahara_output_data_sources.osclients" % CTX)
    def test_setup_and_cleanup_hdfs(self, mock_osclients, mock_cleanup):

        mock_sahara = mock_osclients.Clients(mock.MagicMock()).sahara()
        mock_sahara.data_sources.create.return_value = mock.MagicMock(
            id=42)

        sahara_ctx = sahara_output_data_sources.SaharaOutputDataSources(
            self.context)
        sahara_ctx.generate_random_name = mock.Mock()

        output_ds_crete_calls = []

        for i in range(self.tenants_num):
            output_ds_crete_calls.append(mock.call(
                name=sahara_ctx.generate_random_name.return_value,
                description="",
                data_source_type="hdfs",
                url="hdfs://test_host/"))

        sahara_ctx.setup()

        mock_sahara.data_sources.create.assert_has_calls(
            output_ds_crete_calls)

        sahara_ctx.cleanup()

        mock_cleanup.assert_called_once_with(
            names=["sahara.data_sources"],
            users=self.context["users"])

    @mock.patch("%s.sahara_output_data_sources.osclients" % CTX)
    def test_setup_inputs_swift(self, mock_osclients):
        mock_sahara = mock_osclients.Clients(mock.MagicMock()).sahara()

        self.context.update({
            "config": {
                "users": {
                    "tenants": self.tenants_num,
                    "users_per_tenant": self.users_per_tenant,
                },
                "sahara_output_data_sources": {
                    "output_type": "swift",
                    "output_url_prefix": "rally",
                },
            },
            "admin": {"credential": mock.MagicMock()},
            "task": mock.MagicMock(),
            "users": self.users_key,
            "tenants": self.tenants
        })

        sahara_ctx = sahara_output_data_sources.SaharaOutputDataSources(
            self.context)
        sahara_ctx.generate_random_name = mock.Mock(return_value="random_name")

        output_ds_crete_calls = []
        for i in range(self.tenants_num):
            output_ds_crete_calls.append(mock.call(
                name="random_name",
                description="",
                data_source_type="swift",
                url="swift://random_name.sahara/",
                credential_user="user",
                credential_pass="passwd"
            ))

        sahara_ctx.setup()

        mock_sahara.data_sources.create.assert_has_calls(
            output_ds_crete_calls)

        sahara_ctx.cleanup()



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import copy

import jsonschema
import six

from rally.common import logging
from rally.common.plugin import plugin
from rally.common import utils
from rally import exceptions
from rally.task import functional

LOG = logging.getLogger(__name__)


def configure(name, order, hidden=False):
    """Context class wrapper.

    Each context class has to be wrapped by configure() wrapper. It
    sets essential configuration of context classes. Actually this wrapper just
    adds attributes to the class.

    :param name: Name of the class, used in the input task
    :param order: As far as we can use multiple context classes that sometimes
                  depend on each other we have to specify order of execution.
                  Contexts with smaller order are run first
    :param hidden: If it is true you won't be able to specify context via
                   task config
    """
    def wrapper(cls):
        cls = plugin.configure(name=name)(cls)
        cls._meta_set("order", order)
        cls._meta_set("hidden", hidden)
        return cls

    return wrapper


@configure(name="base_context", order=0, hidden=True)
class Context(plugin.Plugin, functional.FunctionalMixin,
              utils.RandomNameGeneratorMixin):
    """This class is a factory for context classes.

    Every context class should be a subclass of this class and implement
    2 abstract methods: setup() and cleanup()

    It covers:
        1) proper setting up of context config
        2) Auto discovering & get by name
        3) Validation by CONFIG_SCHEMA
        4) Order of context creation

    """
    RESOURCE_NAME_FORMAT = "c_rally_XXXXXXXX_XXXXXXXX"

    CONFIG_SCHEMA = {}

    def __init__(self, ctx):
        self.config = ctx.get("config", {}).get(self.get_name(), {})
        if hasattr(self, "DEFAULT_CONFIG"):
            for key, value in self.DEFAULT_CONFIG.items():
                self.config.setdefault(key, value)
        self.context = ctx
        self.task = self.context.get("task", {})

    def __lt__(self, other):
        return self.get_order() < other.get_order()

    def __gt__(self, other):
        return self.get_order() > other.get_order()

    def __eq__(self, other):
        return self.get_order() == other.get_order()

    @classmethod
    def validate(cls, config, non_hidden=False):
        # TODO(boris-42): This is going to be replaced with common validation
        #                 mechanism (generalization of scenario validation)
        if non_hidden and cls._meta_get("hidden"):
            raise exceptions.PluginNotFound(name=cls.get_name(),
                                            namespace="context")
        jsonschema.validate(config, cls.CONFIG_SCHEMA)

    @classmethod
    def get_order(cls):
        return cls._meta_get("order")

    @abc.abstractmethod
    def setup(self):
        """Prepare environment for test.

        This method is executed only once before load generation.

        self.config contains input arguments of this context
        self.context contains information that will be passed to scenario

        The goal of this method is to perform all operation to prepare
        environment and store information to self.context that is required
        by scenario.
        """

    @abc.abstractmethod
    def cleanup(self):
        """Clean up environment after load generation.

        This method is run once after load generation is done to cleanup
        environment.

        self.config contains input arguments of this context
        self.context contains information that was passed to scenario
        """

    def map_for_scenario(self, context_obj):
        """Maps global context to context that is used by scenario.

        This method is run each time to create scenario context which is
        actually just sub set of global context.

        One of sample where this method is useful is users context.
        We have set of users and tenants but each scenario should have access
        to context of single user in single tenant.
        """
        return context_obj

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.cleanup()


class ContextManager(object):
    """Create context environment and run method inside it."""

    def __init__(self, context_obj):
        self._visited = []
        self.context_obj = context_obj

    @staticmethod
    def validate(ctx, non_hidden=False):
        for name, config in six.iteritems(ctx):
            Context.get(name).validate(config, non_hidden=non_hidden)

    def _get_sorted_context_lst(self):
        ctxlst = map(Context.get, self.context_obj["config"])
        return sorted(map(lambda ctx: ctx(self.context_obj), ctxlst))

    def setup(self):
        """Creates benchmark environment from config."""

        self._visited = []
        for ctx in self._get_sorted_context_lst():
            self._visited.append(ctx)
            ctx.setup()

        return self.context_obj

    def cleanup(self):
        """Destroys benchmark environment."""

        ctxlst = self._visited or self._get_sorted_context_lst()
        for ctx in ctxlst[::-1]:
            try:
                ctx.cleanup()
            except Exception as e:
                LOG.error("Context %s failed during cleanup." % ctx.get_name())
                LOG.exception(e)

    def map_for_scenario(self):
        """Returns scenario's specific context from full context.

        Each context class is able to map context object data as it want.
        E.g. users context will pick one user and it's tenant for each
        scenario run.

        This method iterates over all context classes used in task
        and performs transformation of each of them to full context, order of
        transformation is the same as order of context creation.
        """
        # NOTE(boris-42): Original context_obj is read only and should not
        #                 be modified
        context_obj = copy.deepcopy(self.context_obj)

        for ctx in self._get_sorted_context_lst():
            context_obj = ctx.map_for_scenario(context_obj)

        return context_obj

    def __enter__(self):
        try:
            self.setup()
        except Exception:
            self.cleanup()
            raise

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.cleanup()



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally.common.plugin import plugin
from rally import exceptions
from tests.unit import test


class PluginModuleTestCase(test.TestCase):

    def test_deprecated_func(self):

        @plugin.deprecated("some", "0.0.1")
        @plugin.configure(name="deprecated_func_plugin_test")
        @plugin.from_func()
        def func():
            return 42

        self.assertEqual("deprecated_func_plugin_test", func.get_name())
        self.assertEqual({"reason": "some", "rally_version": "0.0.1"},
                         func.is_deprecated())
        self.assertEqual(42, func())

    def test_configure(self):

        @plugin.configure(name="configure_func_plugin_test")
        @plugin.from_func()
        def func(a):
            return a

        self.assertEqual("configure_func_plugin_test", func.get_name())
        self.assertEqual(42, func(42))

    def test_deprecated_cls(self):

        @plugin.deprecated("God why?", "0.0.2")
        @plugin.configure(name="deprecated_class_plugin_check")
        class MyPlugin(plugin.Plugin):
            pass

        self.assertEqual({"reason": "God why?", "rally_version": "0.0.2"},
                         MyPlugin.is_deprecated())

    def test_configure_cls(self):

        @plugin.configure(name="get_name_class_plugin")
        class MyPlugin(plugin.Plugin):
            pass

        self.assertEqual("get_name_class_plugin", MyPlugin.get_name())

    def test_from_func(self):

        @plugin.from_func()
        def func():
            return 42

        missing = [field for field in set(dir(plugin.Plugin)) - set(dir(func))
                   if not field.startswith("__")]
        self.assertEqual([], missing)
        self.assertTrue(issubclass(func._plugin, plugin.Plugin))
        self.assertEqual(42, func())

    def test_from_func_with_basecls(self):

        class FakeFuncBasePlugin(plugin.Plugin):
            pass

        @plugin.from_func(FakeFuncBasePlugin)
        def func():
            return 43

        self.assertTrue(issubclass(func._plugin, FakeFuncBasePlugin))
        self.assertEqual(43, func())

    def test_from_func_with_bad_basecls(self):

        class FakeFuncBasePlugin(object):
            pass

        self.assertRaises(TypeError,
                          plugin.from_func, FakeFuncBasePlugin)


@plugin.configure(name="test_base_plugin")
class BasePlugin(plugin.Plugin):
    pass


@plugin.configure(name="test_some_plugin")
class SomePlugin(BasePlugin):
    pass


@plugin.deprecated("some_reason", "0.1.1")
@plugin.configure(name="test_deprecated_plugin")
class DeprecatedPlugin(BasePlugin):
    pass


class NotInitedPlugin(BasePlugin):
    pass


class PluginTestCase(test.TestCase):

    def test_unregister(self):

        @plugin.configure(name="test_some_temp_plugin")
        class SomeTempPlugin(BasePlugin):
            pass

        SomeTempPlugin.unregister()
        self.assertRaises(exceptions.PluginNotFound,
                          BasePlugin.get, "test_some_temp_plugin")

    def test_get(self):
        self.assertEqual(SomePlugin,
                         BasePlugin.get("test_some_plugin"))

    def test_get_not_found(self):
        self.assertRaises(exceptions.PluginNotFound,
                          BasePlugin.get, "non_existing")

    def test_get_multple_found(self):

        @plugin.configure("test_2_plugins_with_same_name")
        class A(plugin.Plugin):
            pass

        class B(plugin.Plugin):
            pass

        self.assertRaises(exceptions.PluginWithSuchNameExists,
                          plugin.configure("test_2_plugins_with_same_name"), B)

    def test_get_name(self):
        self.assertEqual("test_some_plugin", SomePlugin.get_name())

    def test_get_all(self):
        self.assertEqual(set([SomePlugin, DeprecatedPlugin]),
                         set(BasePlugin.get_all()))
        self.assertEqual([], SomePlugin.get_all())

    def test_is_deprecated(self):
        self.assertFalse(SomePlugin.is_deprecated())
        self.assertEqual(DeprecatedPlugin.is_deprecated(),
                         {"reason": "some_reason", "rally_version": "0.1.1"})



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml

from rally.plugins.openstack import scenario
from rally.task import atomic


class MistralScenario(scenario.OpenStackScenario):
    """Base class for Mistral scenarios with basic atomic actions."""

    @atomic.action_timer("mistral.list_workbooks")
    def _list_workbooks(self):
        """Gets list of existing workbooks."""
        return self.clients("mistral").workbooks.list()

    @atomic.action_timer("mistral.create_workbook")
    def _create_workbook(self, definition):
        """Create a new workbook.

        :param definition: workbook description in string
                           (yaml string) format
        :returns: workbook object
        """
        definition = yaml.safe_load(definition)
        definition["name"] = self.generate_random_name()
        definition = yaml.safe_dump(definition)

        return self.clients("mistral").workbooks.create(definition)

    @atomic.action_timer("mistral.delete_workbook")
    def _delete_workbook(self, wb_name):
        """Delete the given workbook.

        :param wb_name: the name of workbook that would be deleted.
        """
        self.clients("mistral").workbooks.delete(wb_name)



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" CLI interface for Rally. """

from __future__ import print_function

import sys

from rally.cli import cliutils
from rally.cli import envutils
from rally.common import db


class DBCommands(object):
    """Commands for DB management."""

    def recreate(self):
        """Drop and create Rally database."""
        db.db_drop()
        db.db_create()
        envutils.clear_env()


def main():
    categories = {"db": DBCommands}
    return cliutils.run(sys.argv, categories)


if __name__ == "__main__":
    sys.exit(main())



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally.plugins.openstack.scenarios.fuel import utils
from tests.unit import test


UTILS = "rally.plugins.openstack.scenarios.fuel.utils."


class ModuleTestCase(test.TestCase):

    @mock.patch(UTILS + "six")
    @mock.patch(UTILS + "FuelClient", return_value="fuel_client")
    def test_fuel(self, mock_fuel_client, mock_six):
        mock_six.moves.urllib.parse.urlparse().hostname = "foo_host"
        client = utils.Fuel(
            mock.Mock(username="foo_user", password="foo_pass"),
            {}, {}).create_client()
        mock_fuel_client.assert_called_once_with(
            version="v1", server_address="foo_host", server_port=8000,
            username="foo_user", password="foo_pass")
        self.assertEqual("fuel_client", client)


class FuelEnvTestCase(test.TestCase):

    def test___init__(self):
        env = utils.FuelEnvManager("some_client")
        self.assertEqual("some_client", env.client)

    def test_get(self):
        client = mock.Mock()
        fenv = utils.FuelEnvManager(client)
        result = fenv.get("some_id")
        client.get_by_id.assert_called_once_with("some_id")
        self.assertEqual(result, client.get_by_id("some_id"))
        client.get_by_id.side_effect = BaseException
        self.assertIsNone(fenv.get("some_id"))

    def test_list(self):
        client = mock.Mock()
        envs = [
            {"name": "one"},
            {"name": "two"},
            {"name": "three"}]
        client.get_all.return_value = envs
        fenv = utils.FuelEnvManager(client)
        self.assertEqual(envs, fenv.list())

    def test_list_exception(self):
        client = mock.Mock()
        client.get_all = mock.Mock(side_effect=SystemExit)
        fenv = utils.FuelEnvManager(client)
        self.assertRaises(RuntimeError, fenv.list)

    def test_create(self):
        client = mock.Mock()
        client.create.return_value = "env"
        fenv = utils.FuelEnvManager(client)
        kwargs = {"release_id": 42, "network_provider": "testprov",
                  "deployment_mode": "some_mode", "net_segment_type": "bar"}
        self.assertEqual("env", fenv.create("some_env", **kwargs))
        client.create.assert_called_once_with("some_env", 42, "testprov",
                                              "some_mode", "bar")
        client.create.side_effect = SystemExit
        self.assertRaises(RuntimeError, fenv.create, "some_env", **kwargs)

    def test_create_env_not_returned(self):
        client = mock.Mock()
        client.create.return_value = None
        kwargs = {"release_id": 42, "network_provider": "testprov",
                  "deployment_mode": "some_mode", "net_segment_type": "bar"}
        fenv = utils.FuelEnvManager(client)
        self.assertRaises(RuntimeError, fenv.create, "some_env", **kwargs)

    @mock.patch(UTILS + "scenario.OpenStackScenario")
    def test_delete(self, mock_open_stack_scenario):
        mock_open_stack_scenario.RESOURCE_NAME_PREFIX = ""
        envs = [{"id": "some_one", "name": "one"}]
        client = mock.Mock()
        client.get_all.return_value = envs
        client.delete_by_id.side_effect = SystemExit

        fenv = utils.FuelEnvManager(client)
        self.assertRaises(RuntimeError, fenv.delete, "some_one", retries=2)
        self.assertEqual(3, len(client.delete_by_id.mock_calls))

    @mock.patch(UTILS + "scenario.OpenStackScenario")
    def test_delete_error(self, mock_open_stack_scenario):
        mock_open_stack_scenario.RESOURCE_NAME_PREFIX = ""
        envs = [{"id": "some_one", "name": "one"}]
        client = mock.Mock()
        client.delete_by_id.side_effect = SystemExit
        client.get_all.return_value = envs

        fenv = utils.FuelEnvManager(client)
        self.assertRaises(RuntimeError, fenv.delete, "some_one", retries=1)
        self.assertEqual(2, len(client.delete_by_id.mock_calls))


class FuelClientTestCase(test.TestCase):

    @mock.patch(UTILS + "FuelEnvManager")
    @mock.patch(UTILS + "os")
    def test___init__(self, mock_os, mock_fuel_env_manager):
        mock_os.environ = {}
        mock_fuelclient = mock.Mock(get_client=lambda *args, **kw: [args, kw])
        with mock.patch.dict("sys.modules", {"fuelclient": mock_fuelclient}):
            client = utils.FuelClient(version="foo_version",
                                      server_address="foo_address",
                                      server_port=1234,
                                      username="foo_user",
                                      password="foo_pass")
            expected_environ = {"KEYSTONE_PASS": "foo_pass",
                                "KEYSTONE_USER": "foo_user",
                                "LISTEN_PORT": "1234",
                                "SERVER_ADDRESS": "foo_address"}
            self.assertEqual(expected_environ, mock_os.environ)
            self.assertEqual(mock_fuel_env_manager.return_value,
                             client.environment)
            self.assertEqual([("node",), {"version": "foo_version"}],
                             client.node)
            self.assertEqual([("task",), {"version": "foo_version"}],
                             client.task)
            mock_fuel_env_manager.assert_called_once_with(
                [("environment",),
                 {"version": "foo_version"}])


class FuelScenarioTestCase(test.ScenarioTestCase):

    def test__list_environments(self):
        scenario = utils.FuelScenario(self.context)
        self.assertEqual(
            scenario._list_environments(),
            self.admin_clients("fuel").environment.list.return_value)
        self.admin_clients("fuel").environment.list.assert_called_once_with()
        self._test_atomic_action_timer(scenario.atomic_actions(),
                                       "fuel.list_environments")

    def test__create_environment(self):
        self.admin_clients("fuel").environment.create.return_value = {"id": 42}

        fuel_scenario = utils.FuelScenario()
        fuel_scenario.admin_clients = self.admin_clients

        fuel_scenario.generate_random_name = mock.Mock()
        result = fuel_scenario._create_environment()
        self.assertEqual(
            self.admin_clients("fuel").environment.create.return_value["id"],
            result)
        tmp_mck = self.admin_clients("fuel").environment.create
        tmp_mck.assert_called_once_with(
            fuel_scenario.generate_random_name.return_value, 1, "neutron",
            "ha_compact", "vlan")

    def test__delete_environment(self):
        fuel_scenario = utils.FuelScenario()

        fuel_scenario.admin_clients = self.admin_clients
        fuel_scenario._delete_environment(42, 33)
        tmp_mock = fuel_scenario.admin_clients("fuel")
        tmp_mock.environment.delete.assert_called_once_with(42, 33)

    def test__add_nodes(self):
        fscen = utils.FuelScenario()
        fscen.admin_clients = mock.Mock()
        fscen._add_node("1", ["42"], node_roles=["some_role"])
        tmp_mock = fscen.admin_clients.return_value.environment.client
        tmp_mock.add_nodes.assert_called_once_with("1", ["42"], ["some_role"])

    def test__add_nodes_error(self):
        fscen = utils.FuelScenario()
        fscen.admin_clients = mock.Mock()
        tmp_mock = fscen.admin_clients.return_value.environment.client
        tmp_mock.add_nodes.side_effect = BaseException
        self.assertRaises(RuntimeError, fscen._add_node, "1", "42",
                          node_roles="some_role")

    @mock.patch(UTILS + "FuelClient")
    def test__remove_nodes(self, mock_fuel_client):
        mock_tmp = mock_fuel_client.fuelclient_module.objects
        mock_env = mock_tmp.environment.Environment
        mock_env.return_value = mock.Mock()
        fscen = utils.FuelScenario()
        fscen._remove_node("1", "2")
        mock_env.assert_called_once_with("1")
        mock_env.return_value.unassign.assert_called_once_with(["2"])

    @mock.patch(UTILS + "FuelClient")
    def test__remove_nodes_error(self, mock_fuel_client):
        mock_tmp = mock_fuel_client.fuelclient_module.objects
        mock_env = mock_tmp.environment.Environment
        mock_env.return_value = mock.Mock()
        mock_env.return_value.unassign.side_effect = BaseException
        fscen = utils.FuelScenario()
        self.assertRaises(RuntimeError, fscen._remove_node, "1", "2")

    def test__list_node_ids(self):
        fscen = utils.FuelScenario()
        fscen.admin_clients = mock.Mock()
        fscen.admin_clients.return_value.node.get_all.return_value = [
            {"id": "id1"}, {"id": "id2"}]
        res = fscen._list_node_ids("env")
        self.assertEqual(["id1", "id2"], res)
        tmp_mock = fscen.admin_clients.return_value.node.get_all
        tmp_mock.assert_called_once_with(environment_id="env")

    def test__node_is_assigned(self):
        fscen = utils.FuelScenario()
        fscen.admin_clients = mock.Mock()
        fscen.admin_clients.return_value.node.get_by_id.return_value = {
            "id": "id1", "cluster": "some_id"}
        self.assertTrue(fscen._node_is_assigned("id1"))
        fscen.admin_clients.return_value.node.get_by_id.return_value[
            "cluster"] = ""
        self.assertFalse(fscen._node_is_assigned("id2"))

    @mock.patch(UTILS + "FuelScenario._node_is_assigned", return_value=False)
    @mock.patch(UTILS + "FuelScenario._list_node_ids",
                return_value=["id1", "id2"])
    def test__get_free_node_id(self, mock__list_node_ids,
                               mock__node_is_assigned):
        node_id = utils.FuelScenario()._get_free_node_id()
        self.assertIn(node_id, mock__list_node_ids.return_value)

    @mock.patch(UTILS + "FuelScenario._node_is_assigned", return_value=True)
    @mock.patch(UTILS + "FuelScenario._list_node_ids",
                return_value=["id1", "id2"])
    def test__get_free_node_id_exception(self, mock__list_node_ids,
                                         mock__node_is_assigned):
        self.assertRaises(RuntimeError,
                          utils.FuelScenario()._get_free_node_id)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally import consts
from rally.plugins.openstack import scenario
from rally.plugins.openstack.scenarios.ceilometer import utils as ceiloutils
from rally.task import validation


class CeilometerSamples(ceiloutils.CeilometerScenario):
    """Benchmark scenarios for Ceilometer Samples API."""

    @validation.required_services(consts.Service.CEILOMETER)
    @validation.required_openstack(users=True)
    @scenario.configure()
    def list_matched_samples(self, filter_by_resource_id=False,
                             filter_by_project_id=False,
                             filter_by_user_id=False,
                             metadata_query=None, limit=None):
        """Get list of samples that matched fields from context and args.

        :param filter_by_user_id: flag for query by user_id
        :param filter_by_project_id: flag for query by project_id
        :param filter_by_resource_id: flag for query by resource_id
        :param metadata_query: dict with metadata fields and values for query
        :param limit: count of samples in response
        """
        query = self._make_general_query(filter_by_project_id,
                                         filter_by_user_id,
                                         filter_by_resource_id,
                                         metadata_query)
        self._list_samples(query, limit)

    @validation.required_services(consts.Service.CEILOMETER)
    @validation.required_openstack(users=True)
    @scenario.configure()
    def list_samples(self, metadata_query=None, limit=None):
        """Fetch all available queries for list sample request.

        :param metadata_query: dict with metadata fields and values for query
        :param limit: count of samples in response
        """
        self.list_matched_samples(filter_by_project_id=True)
        self.list_matched_samples(filter_by_user_id=True)
        self.list_matched_samples(filter_by_resource_id=True)
        if metadata_query:
            self.list_matched_samples(metadata_query=metadata_query)
        if limit:
            self.list_matched_samples(limit=limit)



# Copyright 2014: Rackspace UK
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from rally.common import logging
from rally import consts
from rally import exceptions
from rally.plugins.openstack import scenario
from rally.plugins.openstack.scenarios.vm import utils as vm_utils
from rally.task import types
from rally.task import validation


class VMTasks(vm_utils.VMScenario):
    """Benchmark scenarios that are to be run inside VM instances."""

    def __init__(self, *args, **kwargs):
        super(VMTasks, self).__init__(*args, **kwargs)

    @types.set(image=types.ImageResourceType,
               flavor=types.FlavorResourceType)
    @validation.image_valid_on_flavor("flavor", "image")
    @logging.log_deprecated_args("Use `command' argument instead", "0.0.5",
                                 ("script", "interpreter"), once=True)
    @validation.file_exists("script", required=False)
    @validation.valid_command("command", required=False)
    @validation.number("port", minval=1, maxval=65535, nullable=True,
                       integer_only=True)
    @validation.external_network_exists("floating_network")
    @validation.required_services(consts.Service.NOVA, consts.Service.CINDER)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"cleanup": ["nova", "cinder"],
                                 "keypair": {}, "allow_ssh": {}})
    def boot_runcommand_delete(self, image, flavor,
                               username,
                               password=None,
                               script=None,
                               interpreter=None,
                               command=None,
                               volume_args=None,
                               floating_network=None,
                               port=22,
                               use_floating_ip=True,
                               force_delete=False,
                               wait_for_ping=True,
                               **kwargs):
        """Boot a server, run a script that outputs JSON, delete the server.

        Example Script in samples/tasks/support/instance_dd_test.sh

        :param image: glance image name to use for the vm
        :param flavor: VM flavor name
        :param username: ssh username on server, str
        :param password: Password on SSH authentication
        :param script: DEPRECATED. Use `command' instead. Script to run on
            server, must output JSON mapping metric names to values (see the
            sample script below)
        :param interpreter: DEPRECATED. Use `command' instead. server's
            interpreter to run the script
        :param command: Command-specifying dictionary that either specifies
            remote command path via `remote_path' (can be uploaded from a
            local file specified by `local_path`), an inline script via
            `script_inline' or a local script file path using `script_file'.
            Both `script_file' and `local_path' are checked to be accessible
            by the `file_exists' validator code.

            The `script_inline' and `script_file' both require an `interpreter'
            value to specify the interpreter script should be run with.

            Note that any of `interpreter' and `remote_path' can be an array
            prefixed with environment variables and suffixed with args for
            the `interpreter' command. `remote_path's last component must be
            a path to a command to execute (also upload destination if a
            `local_path' is given). Uploading an interpreter is possible
            but requires that `remote_path' and `interpreter' path do match.


            Examples::

                # Run a `local_script.pl' file sending it to a remote
                # Perl interpreter
                command = {
                    "script_file": "local_script.pl",
                    "interpreter": "/usr/bin/perl"
                }

                # Run an inline script sending it to a remote interpreter
                command = {
                    "script_inline": "echo 'Hello, World!'",
                    "interpreter": "/bin/sh"
                }

                # Run a remote command
                command = {
                    "remote_path": "/bin/false"
                }

                # Copy a local command and run it
                command = {
                    "remote_path": "/usr/local/bin/fio",
                    "local_path": "/home/foobar/myfiodir/bin/fio"
                }

                # Copy a local command and run it with environment variable
                command = {
                    "remote_path": ["HOME=/root", "/usr/local/bin/fio"],
                    "local_path": "/home/foobar/myfiodir/bin/fio"
                }

                # Run an inline script sending it to a remote interpreter
                command = {
                    "script_inline": "echo \"Hello, ${NAME:-World}\"",
                    "interpreter": ["NAME=Earth", "/bin/sh"]
                }

                # Run an inline script sending it to an uploaded remote
                # interpreter
                command = {
                    "script_inline": "echo \"Hello, ${NAME:-World}\"",
                    "interpreter": ["NAME=Earth", "/tmp/sh"],
                    "remote_path": "/tmp/sh",
                    "local_path": "/home/user/work/cve/sh-1.0/bin/sh"
                }


        :param volume_args: volume args for booting server from volume
        :param floating_network: external network name, for floating ip
        :param port: ssh port for SSH connection
        :param use_floating_ip: bool, floating or fixed IP for SSH connection
        :param force_delete: whether to use force_delete for servers
        :param wait_for_ping: whether to check connectivity on server creation
        :param **kwargs: extra arguments for booting the server
        :returns: dictionary with keys `data' and `errors':
                  data: dict, JSON output from the script
                  errors: str, raw data from the script's stderr stream
        """

        if command is None and script and interpreter:
            command = {"script_file": script, "interpreter": interpreter}

        if volume_args:
            volume = self._create_volume(volume_args["size"], imageRef=None)
            kwargs["block_device_mapping"] = {"vdrally": "%s:::1" % volume.id}

        server, fip = self._boot_server_with_fip(
            image, flavor, use_floating_ip=use_floating_ip,
            floating_network=floating_network,
            key_name=self.context["user"]["keypair"]["name"],
            **kwargs)
        try:
            if wait_for_ping:
                self._wait_for_ping(fip["ip"])

            code, out, err = self._run_command(
                fip["ip"], port, username, password, command=command)
            if code:
                raise exceptions.ScriptError(
                    "Error running command %(command)s. "
                    "Error %(code)s: %(error)s" % {
                        "command": command, "code": code, "error": err})

            try:
                data = json.loads(out)
            except ValueError as e:
                raise exceptions.ScriptError(
                    "Command %(command)s has not output valid JSON: %(error)s."
                    " Output: %(output)s" % {
                        "command": command, "error": str(e), "output": out})
        finally:
            self._delete_server_with_fip(server, fip,
                                         force_delete=force_delete)

        return {"data": data, "errors": err}

    @types.set(image=types.ImageResourceType,
               flavor=types.FlavorResourceType)
    @validation.number("port", minval=1, maxval=65535, nullable=True,
                       integer_only=True)
    @validation.valid_command("command")
    @validation.external_network_exists("floating_network")
    @validation.required_services(consts.Service.NOVA, consts.Service.CINDER)
    @validation.required_openstack(users=True)
    @validation.required_contexts("image_command_customizer")
    @scenario.configure(context={"cleanup": ["nova", "cinder"],
                                 "keypair": {}, "allow_ssh": {}})
    def boot_runcommand_delete_custom_image(self, **kwargs):
        """Boot a server from a custom image, run a command that outputs JSON.

        Example Script in rally-jobs/extra/install_benchmark.sh
        """

        return self.boot_runcommand_delete(
            image=self.context["tenant"]["custom_image"]["id"], **kwargs)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally.common import logging
from rally import consts
from rally.plugins.openstack import scenario
from rally.plugins.openstack.scenarios.ceilometer import utils
from rally.task import validation


class CeilometerStats(utils.CeilometerScenario):
    """Benchmark scenarios for Ceilometer Stats API."""

    @logging.log_deprecated("Use 'get_stats' method, now samples are created"
                            "in context", "0.1.2")
    @validation.required_services(consts.Service.CEILOMETER)
    @validation.required_openstack(users=True)
    @scenario.configure()
    def create_meter_and_get_stats(self, **kwargs):
        """Create a meter and fetch its statistics.

        Meter is first created and then statistics is fetched for the same
        using GET /v2/meters/(meter_name)/statistics.

        :param kwargs: contains optional arguments to create a meter
        """
        meter = self._create_meter(**kwargs)
        self._get_stats(meter.counter_name)

    @validation.required_services(consts.Service.CEILOMETER)
    @validation.required_openstack(users=True)
    @scenario.configure()
    def get_stats(self, meter_name, filter_by_user_id=False,
                  filter_by_project_id=False, filter_by_resource_id=False,
                  metadata_query=None, period=None, groupby=None,
                  aggregates=None):
        """Fetch statistics for certain meter.

        Statistics is fetched for the using
        GET /v2/meters/(meter_name)/statistics.

        :param meter_name: meter to take statistic for
        :param filter_by_user_id: flag for query by user_id
        :param filter_by_project_id: flag for query by project_id
        :param filter_by_resource_id: flag for query by resource_id
        :param metadata_query: dict with metadata fields and values for query
        :param period: the length of the time range covered by these stats
        :param groupby: the fields used to group the samples
        :param aggregates: name of function for samples aggregation

        :returns: list of statistics data
        """
        query = self._make_general_query(filter_by_project_id,
                                         filter_by_user_id,
                                         filter_by_resource_id,
                                         metadata_query)
        self._get_stats(meter_name, query, period, groupby, aggregates)



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import zipfile

from rally.common import fileutils
from rally.common.i18n import _
from rally.common.i18n import _LE
from rally.common import logging
from rally.common import utils
from rally import consts
from rally import exceptions
from rally import osclients
from rally.plugins.openstack.context.cleanup import manager as resource_manager
from rally.task import context


LOG = logging.getLogger(__name__)


@context.configure(name="murano_packages", order=401)
class PackageGenerator(context.Context):
    """Context class for uploading applications for murano."""

    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA,
        "properties": {
            "app_package": {
                "type": "string",
            }
        },
        "required": ["app_package"],
        "additionalProperties": False
    }

    @logging.log_task_wrapper(LOG.info, _("Enter context: `Murano packages`"))
    def setup(self):
        is_config_app_dir = False
        pckg_path = os.path.expanduser(self.config["app_package"])
        if zipfile.is_zipfile(pckg_path):
            zip_name = pckg_path
        elif os.path.isdir(pckg_path):
            is_config_app_dir = True
            zip_name = fileutils.pack_dir(pckg_path)
        else:
            msg = (_LE("There is no zip archive or directory by this path:"
                       " %s") % pckg_path)
            raise exceptions.ContextSetupFailure(msg=msg,
                                                 ctx_name=self.get_name())

        for user, tenant_id in utils.iterate_per_tenants(
                self.context["users"]):
            clients = osclients.Clients(user["credential"])
            self.context["tenants"][tenant_id]["packages"] = []
            if is_config_app_dir:
                self.context["tenants"][tenant_id]["murano_ctx"] = zip_name
            package = clients.murano().packages.create(
                {"categories": ["Web"], "tags": ["tag"]},
                {"file": open(zip_name)})

            self.context["tenants"][tenant_id]["packages"].append(package)

    @logging.log_task_wrapper(LOG.info, _("Exit context: `Murano packages`"))
    def cleanup(self):
        resource_manager.cleanup(names=["murano.packages"],
                                 users=self.context.get("users", []))



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_config import cfg
import requests

from rally.common import logging
from rally import exceptions
from rally.plugins.openstack import scenario
from rally.task import atomic
from rally.task import utils

LOG = logging.getLogger(__name__)

HEAT_BENCHMARK_OPTS = [
    cfg.FloatOpt("heat_stack_create_prepoll_delay",
                 default=2.0,
                 help="Time(in sec) to sleep after creating a resource before "
                      "polling for it status."),
    cfg.FloatOpt("heat_stack_create_timeout",
                 default=3600.0,
                 help="Time(in sec) to wait for heat stack to be created."),
    cfg.FloatOpt("heat_stack_create_poll_interval",
                 default=1.0,
                 help="Time interval(in sec) between checks when waiting for "
                      "stack creation."),
    cfg.FloatOpt("heat_stack_delete_timeout",
                 default=3600.0,
                 help="Time(in sec) to wait for heat stack to be deleted."),
    cfg.FloatOpt("heat_stack_delete_poll_interval",
                 default=1.0,
                 help="Time interval(in sec) between checks when waiting for "
                      "stack deletion."),
    cfg.FloatOpt("heat_stack_check_timeout",
                 default=3600.0,
                 help="Time(in sec) to wait for stack to be checked."),
    cfg.FloatOpt("heat_stack_check_poll_interval",
                 default=1.0,
                 help="Time interval(in sec) between checks when waiting for "
                      "stack checking."),
    cfg.FloatOpt("heat_stack_update_prepoll_delay",
                 default=2.0,
                 help="Time(in sec) to sleep after updating a resource before "
                      "polling for it status."),
    cfg.FloatOpt("heat_stack_update_timeout",
                 default=3600.0,
                 help="Time(in sec) to wait for stack to be updated."),
    cfg.FloatOpt("heat_stack_update_poll_interval",
                 default=1.0,
                 help="Time interval(in sec) between checks when waiting for "
                      "stack update."),
    cfg.FloatOpt("heat_stack_suspend_timeout",
                 default=3600.0,
                 help="Time(in sec) to wait for stack to be suspended."),
    cfg.FloatOpt("heat_stack_suspend_poll_interval",
                 default=1.0,
                 help="Time interval(in sec) between checks when waiting for "
                      "stack suspend."),
    cfg.FloatOpt("heat_stack_resume_timeout",
                 default=3600.0,
                 help="Time(in sec) to wait for stack to be resumed."),
    cfg.FloatOpt("heat_stack_resume_poll_interval",
                 default=1.0,
                 help="Time interval(in sec) between checks when waiting for "
                      "stack resume."),
    cfg.FloatOpt("heat_stack_snapshot_timeout",
                 default=3600.0,
                 help="Time(in sec) to wait for stack snapshot to "
                      "be created."),
    cfg.FloatOpt("heat_stack_snapshot_poll_interval",
                 default=1.0,
                 help="Time interval(in sec) between checks when waiting for "
                      "stack snapshot to be created."),
    cfg.FloatOpt("heat_stack_restore_timeout",
                 default=3600.0,
                 help="Time(in sec) to wait for stack to be restored from "
                      "snapshot."),
    cfg.FloatOpt("heat_stack_restore_poll_interval",
                 default=1.0,
                 help="Time interval(in sec) between checks when waiting for "
                      "stack to be restored."),
    cfg.FloatOpt("heat_stack_scale_timeout",
                 default=3600.0,
                 help="Time (in sec) to wait for stack to scale up or down."),
    cfg.FloatOpt("heat_stack_scale_poll_interval",
                 default=1.0,
                 help="Time interval (in sec) between checks when waiting for "
                      "a stack to scale up or down."),
]

CONF = cfg.CONF
benchmark_group = cfg.OptGroup(name="benchmark", title="benchmark options")
CONF.register_opts(HEAT_BENCHMARK_OPTS, group=benchmark_group)


class HeatScenario(scenario.OpenStackScenario):
    """Base class for Heat scenarios with basic atomic actions."""

    @atomic.action_timer("heat.list_stacks")
    def _list_stacks(self):
        """Return user stack list."""

        return list(self.clients("heat").stacks.list())

    @atomic.action_timer("heat.create_stack")
    def _create_stack(self, template, parameters=None,
                      files=None, environment=None):
        """Create a new stack.

        :param template: template with stack description.
        :param parameters: template parameters used during stack creation
        :param files: additional files used in template
        :param environment: stack environment definition

        :returns: object of stack
        """
        stack_name = self.generate_random_name()
        kw = {
            "stack_name": stack_name,
            "disable_rollback": True,
            "parameters": parameters or {},
            "template": template,
            "files": files or {},
            "environment": environment or {}
        }

        # heat client returns body instead manager object, so we should
        # get manager object using stack_id
        stack_id = self.clients("heat").stacks.create(**kw)["stack"]["id"]
        stack = self.clients("heat").stacks.get(stack_id)

        time.sleep(CONF.benchmark.heat_stack_create_prepoll_delay)

        stack = utils.wait_for(
            stack,
            ready_statuses=["CREATE_COMPLETE"],
            update_resource=utils.get_from_manager(["CREATE_FAILED"]),
            timeout=CONF.benchmark.heat_stack_create_timeout,
            check_interval=CONF.benchmark.heat_stack_create_poll_interval)

        return stack

    @atomic.action_timer("heat.update_stack")
    def _update_stack(self, stack, template, parameters=None,
                      files=None, environment=None):
        """Update an existing stack

        :param stack: stack that need to be updated
        :param template: Updated template
        :param parameters: template parameters for stack update
        :param files: additional files used in template
        :param environment: stack environment definition

        :returns: object of updated stack
        """

        kw = {
            "stack_name": stack.stack_name,
            "disable_rollback": True,
            "parameters": parameters or {},
            "template": template,
            "files": files or {},
            "environment": environment or {}
        }
        self.clients("heat").stacks.update(stack.id, **kw)

        time.sleep(CONF.benchmark.heat_stack_update_prepoll_delay)
        stack = utils.wait_for(
            stack,
            ready_statuses=["UPDATE_COMPLETE"],
            update_resource=utils.get_from_manager(["UPDATE_FAILED"]),
            timeout=CONF.benchmark.heat_stack_update_timeout,
            check_interval=CONF.benchmark.heat_stack_update_poll_interval)
        return stack

    @atomic.action_timer("heat.check_stack")
    def _check_stack(self, stack):
        """Check given stack.

        Check the stack and stack resources.

        :param stack: stack that needs to be checked
        """
        self.clients("heat").actions.check(stack.id)
        utils.wait_for(
            stack,
            ready_statuses=["CHECK_COMPLETE"],
            update_resource=utils.get_from_manager(["CHECK_FAILED"]),
            timeout=CONF.benchmark.heat_stack_check_timeout,
            check_interval=CONF.benchmark.heat_stack_check_poll_interval)

    @atomic.action_timer("heat.delete_stack")
    def _delete_stack(self, stack):
        """Delete given stack.

        Returns when the stack is actually deleted.

        :param stack: stack object
        """
        stack.delete()
        utils.wait_for_status(
            stack,
            ready_statuses=["deleted"],
            check_deletion=True,
            update_resource=utils.get_from_manager(),
            timeout=CONF.benchmark.heat_stack_delete_timeout,
            check_interval=CONF.benchmark.heat_stack_delete_poll_interval)

    @atomic.action_timer("heat.suspend_stack")
    def _suspend_stack(self, stack):
        """Suspend given stack.

        :param stack: stack that needs to be suspended
        """

        self.clients("heat").actions.suspend(stack.id)
        utils.wait_for(
            stack,
            ready_statuses=["SUSPEND_COMPLETE"],
            update_resource=utils.get_from_manager(
                ["SUSPEND_FAILED"]),
            timeout=CONF.benchmark.heat_stack_suspend_timeout,
            check_interval=CONF.benchmark.heat_stack_suspend_poll_interval)

    @atomic.action_timer("heat.resume_stack")
    def _resume_stack(self, stack):
        """Resume given stack.

        :param stack: stack that needs to be resumed
        """

        self.clients("heat").actions.resume(stack.id)
        utils.wait_for(
            stack,
            ready_statuses=["RESUME_COMPLETE"],
            update_resource=utils.get_from_manager(
                ["RESUME_FAILED"]),
            timeout=CONF.benchmark.heat_stack_resume_timeout,
            check_interval=CONF.benchmark.heat_stack_resume_poll_interval)

    @atomic.action_timer("heat.snapshot_stack")
    def _snapshot_stack(self, stack):
        """Creates a snapshot for given stack.

        :param stack: stack that will be used as base for snapshot
        :returns: snapshot created for given stack
        """
        snapshot = self.clients("heat").stacks.snapshot(
            stack.id)
        utils.wait_for(
            stack,
            ready_statuses=["SNAPSHOT_COMPLETE"],
            update_resource=utils.get_from_manager(
                ["SNAPSHOT_FAILED"]),
            timeout=CONF.benchmark.heat_stack_snapshot_timeout,
            check_interval=CONF.benchmark.heat_stack_snapshot_poll_interval)
        return snapshot

    @atomic.action_timer("heat.restore_stack")
    def _restore_stack(self, stack, snapshot_id):
        """Restores stack from given snapshot.

        :param stack: stack that will be restored from snapshot
        :param snapshot_id: id of given snapshot
        """
        self.clients("heat").stacks.restore(stack.id, snapshot_id)
        utils.wait_for(
            stack,
            ready_statuses=["RESTORE_COMPLETE"],
            update_resource=utils.get_from_manager(
                ["RESTORE_FAILED"]),
            timeout=CONF.benchmark.heat_stack_restore_timeout,
            check_interval=CONF.benchmark.heat_stack_restore_poll_interval
        )

    def _count_instances(self, stack):
        """Count instances in a Heat stack.

        :param stack: stack to count instances in.
        """
        return len([
            r for r in self.clients("heat").resources.list(stack.id,
                                                           nested_depth=1)
            if r.resource_type == "OS::Nova::Server"])

    def _scale_stack(self, stack, output_key, delta):
        """Scale a stack up or down.

        Calls the webhook given in the output value identified by
        'output_key', and waits for the stack size to change by
        'delta'.

        :param stack: stack to scale up or down
        :param output_key: The name of the output to get the URL from
        :param delta: The expected change in number of instances in
                      the stack (signed int)
        """
        num_instances = self._count_instances(stack)
        expected_instances = num_instances + delta
        LOG.debug("Scaling stack %s from %s to %s instances with %s" %
                  (stack.id, num_instances, expected_instances, output_key))
        with atomic.ActionTimer(self, "heat.scale_with_%s" % output_key):
            self._stack_webhook(stack, output_key)
            utils.wait_for(
                stack,
                is_ready=lambda s: (
                    self._count_instances(s) == expected_instances),
                update_resource=utils.get_from_manager(
                    ["UPDATE_FAILED"]),
                timeout=CONF.benchmark.heat_stack_scale_timeout,
                check_interval=CONF.benchmark.heat_stack_scale_poll_interval)

    def _stack_webhook(self, stack, output_key):
        """POST to the URL given in the output value identified by output_key.

        This can be used to scale stacks up and down, for instance.

        :param stack: stack to call a webhook on
        :param output_key: The name of the output to get the URL from
        :raises InvalidConfigException: if the output key is not found
        """
        url = None
        for output in stack.outputs:
            if output["output_key"] == output_key:
                url = output["output_value"]
                break
        else:
            raise exceptions.InvalidConfigException(
                "No output key %(key)s found in stack %(id)s" %
                {"key": output_key, "id": stack.id})

        with atomic.ActionTimer(self, "heat.%s_webhook" % output_key):
            requests.post(url).raise_for_status()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

import six

from rally.common.i18n import _, _LE
from rally.common import logging
from rally import consts
from rally import exceptions
from rally import osclients
from rally.task import context

LOG = logging.getLogger(__name__)


@context.configure(name="api_versions", order=150)
class OpenStackAPIVersions(context.Context):
    """Context for specifying OpenStack clients versions and service types."""

    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA,
        "patternProperties": {
            "^[a-z]+$": {
                "type": "object",
                "properties": {
                    "version": {
                        "anyOf": [{"type": "string"}, {"type": "number"}]
                    },
                    "service_name": {
                        "type": "string"
                    },
                    "service_type": {
                        "type": "string"
                    }
                },
                "additionalProperties": False
            }
        },
        "additionalProperties": False
    }

    def setup(self):
        # FIXME(andreykurilin): move all checks to validate method.

        # use admin only when `service_name` is presented
        admin_clients = osclients.Clients(
            self.context.get("admin", {}).get("credential"))
        clients = osclients.Clients(random.choice(
            self.context["users"])["credential"])
        services = clients.keystone().service_catalog.get_endpoints()
        services_from_admin = None
        for client_name, conf in six.iteritems(self.config):
            if "service_type" in conf and conf["service_type"] not in services:
                raise exceptions.ValidationError(_(
                    "There is no service with '%s' type in your environment.")
                    % conf["service_type"])
            elif "service_name" in conf:
                if not self.context.get("admin", {}).get("credential"):
                    raise exceptions.BenchmarkSetupFailure(_(
                        "Setting 'service_name' is allowed only for 'admin' "
                        "user."))
                if not services_from_admin:
                    services_from_admin = dict(
                        [(s.name, s.type)
                         for s in admin_clients.keystone().services.list()])
                if conf["service_name"] not in services_from_admin:
                    raise exceptions.ValidationError(
                        _("There is no '%s' service in your environment") %
                        conf["service_name"])

                self.context["config"]["api_versions"][client_name][
                    "service_type"] = services_from_admin[conf["service_name"]]

    def cleanup(self):
        # nothing to do here
        pass

    @classmethod
    def validate(cls, config, non_hidden=False):
        super(OpenStackAPIVersions, cls).validate(config,
                                                  non_hidden=non_hidden)
        for client in config:
            client_cls = osclients.OSClient.get(client)
            if ("service_type" in config[client] and
                    "service_name" in config[client]):
                raise exceptions.ValidationError(_LE(
                    "Setting both 'service_type' and 'service_name' properties"
                    " is restricted."))
            try:
                if ("service_type" in config[client] or
                        "service_name" in config[client]):
                    client_cls.is_service_type_configurable()

                if "version" in config[client]:
                    client_cls.validate_version(config[client]["version"])

            except exceptions.RallyException as e:
                raise exceptions.ValidationError(
                    _LE("Invalid settings for '%(client)s': %(error)s") % {
                        "client": client,
                        "error": e.format_message()})



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally.plugins.openstack.scenarios.glance import images
from tests.unit import fakes
from tests.unit import test

GLANCE_IMAGES = "rally.plugins.openstack.scenarios.glance.images.GlanceImages"


class GlanceImagesTestCase(test.ScenarioTestCase):

    @mock.patch(GLANCE_IMAGES + ".generate_random_name")
    @mock.patch(GLANCE_IMAGES + "._list_images")
    @mock.patch(GLANCE_IMAGES + "._create_image")
    def test_create_and_list_image(self, mock__create_image,
                                   mock__list_images,
                                   mock_generate_random_name):
        glance_scenario = images.GlanceImages(self.context)
        mock_generate_random_name.return_value = "test-rally-image"
        glance_scenario.create_and_list_image("cf", "url", "df",
                                              fakearg="f")
        mock__create_image.assert_called_once_with(
            "cf", "url", "df", fakearg="f")
        mock__list_images.assert_called_once_with()

    @mock.patch(GLANCE_IMAGES + "._list_images")
    def test_list_images(self, mock__list_images):
        glance_scenario = images.GlanceImages(self.context)
        glance_scenario.list_images()
        mock__list_images.assert_called_once_with()

    @mock.patch(GLANCE_IMAGES + ".generate_random_name")
    @mock.patch(GLANCE_IMAGES + "._delete_image")
    @mock.patch(GLANCE_IMAGES + "._create_image")
    def test_create_and_delete_image(
            self, mock__create_image, mock__delete_image,
            mock_generate_random_name):
        glance_scenario = images.GlanceImages(self.context)
        fake_image = object()
        mock__create_image.return_value = fake_image
        mock_generate_random_name.return_value = "test-rally-image"
        glance_scenario.create_and_delete_image("cf", "url", "df",
                                                fakearg="f")

        mock__create_image.assert_called_once_with(
            "cf", "url", "df", fakearg="f")
        mock__delete_image.assert_called_once_with(fake_image)

    @mock.patch(GLANCE_IMAGES + "._boot_servers")
    @mock.patch(GLANCE_IMAGES + "._create_image")
    def test_create_image_and_boot_instances(
            self, mock__create_image, mock__boot_servers):
        glance_scenario = images.GlanceImages(self.context)
        fake_image = fakes.FakeImage()
        fake_servers = [mock.Mock() for i in range(5)]
        mock__create_image.return_value = fake_image
        mock__boot_servers.return_value = fake_servers
        kwargs = {"fakearg": "f"}

        glance_scenario.create_image_and_boot_instances("cf", "url",
                                                        "df", "fid",
                                                        5, **kwargs)
        mock__create_image.assert_called_once_with(
            "cf", "url", "df")
        mock__boot_servers.assert_called_once_with(
            "image-id-0", "fid", 5, **kwargs)



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

import ddt
import mock

from rally.task.processing import plot
from tests.unit import test

PLOT = "rally.task.processing.plot."


@ddt.ddt
class PlotTestCase(test.TestCase):

    @mock.patch(PLOT + "charts")
    def test__process_scenario(self, mock_charts):
        for mock_ins, ret in [
                (mock_charts.MainStatsTable, "main_stats"),
                (mock_charts.MainStackedAreaChart, "main_stacked"),
                (mock_charts.AtomicStackedAreaChart, "atomic_stacked"),
                (mock_charts.OutputStackedAreaDeprecatedChart,
                 "output_stacked"),
                (mock_charts.LoadProfileChart, "load_profile"),
                (mock_charts.MainHistogramChart, "main_histogram"),
                (mock_charts.AtomicHistogramChart, "atomic_histogram"),
                (mock_charts.AtomicAvgChart, "atomic_avg")]:
            setattr(mock_ins.return_value.render, "return_value", ret)
        iterations = [
            {"timestamp": i + 2, "error": [],
             "duration": i + 5, "idle_duration": i,
             "output": {"additive": [], "complete": []},
             "atomic_actions": {"foo_action": i + 10}} for i in range(10)]
        data = {"iterations": iterations, "sla": [],
                "key": {"kw": {"runner": {"type": "constant"}},
                        "name": "Foo.bar", "pos": 0},
                "info": {"atomic": {"foo_action": {"max_duration": 19,
                                                   "min_duration": 10}},
                         "full_duration": 40, "load_duration": 32,
                         "iterations_count": 10, "iterations_passed": 10,
                         "max_duration": 14, "min_duration": 5,
                         "output_names": [],
                         "tstamp_end": 25, "tstamp_start": 2}}

        task_data = plot._process_scenario(data, 1)
        self.assertEqual(
            task_data, {
                "cls": "Foo", "met": "bar", "name": "bar [2]", "pos": "1",
                "runner": "constant", "config": json.dumps(
                    {"Foo.bar": [{"runner": {"type": "constant"}}]},
                    indent=2),
                "full_duration": 40, "load_duration": 32,
                "atomic": {"histogram": "atomic_histogram",
                           "iter": "atomic_stacked", "pie": "atomic_avg"},
                "iterations": {"histogram": "main_histogram",
                               "iter": "main_stacked",
                               "pie": [("success", 10), ("errors", 0)]},
                "iterations_count": 10, "errors": [],
                "load_profile": "load_profile",
                "additive_output": [],
                "complete_output": [[], [], [], [], [], [], [], [], [], []],
                "output_errors": [],
                "sla": [], "sla_success": True, "table": "main_stats"})

    @mock.patch(PLOT + "_process_scenario")
    @mock.patch(PLOT + "json.dumps", return_value="json_data")
    def test__process_tasks(self, mock_json_dumps, mock__process_scenario):
        tasks_results = [{"key": {"name": i, "kw": "kw_" + i}}
                         for i in ("a", "b", "c", "b")]
        mock__process_scenario.side_effect = lambda a, b: (
            {"cls": "%s_cls" % a["key"]["name"],
             "name": str(b),
             "met": "dummy",
             "pos": str(b)})
        source, tasks = plot._process_tasks(tasks_results)
        self.assertEqual(source, "json_data")
        mock_json_dumps.assert_called_once_with(
            {"a": ["kw_a"], "b": ["kw_b", "kw_b"], "c": ["kw_c"]},
            sort_keys=True, indent=2)
        self.assertEqual(
            tasks,
            [{"cls": "a_cls", "met": "dummy", "name": "0", "pos": "0"},
             {"cls": "b_cls", "met": "dummy", "name": "0", "pos": "0"},
             {"cls": "b_cls", "met": "dummy", "name": "1", "pos": "1"},
             {"cls": "c_cls", "met": "dummy", "name": "0", "pos": "0"}])

    @ddt.data({},
              {"include_libs": True},
              {"include_libs": False})
    @ddt.unpack
    @mock.patch(PLOT + "_process_tasks")
    @mock.patch(PLOT + "objects")
    @mock.patch(PLOT + "ui_utils.get_template")
    @mock.patch(PLOT + "json.dumps", side_effect=lambda s: "json_" + s)
    def test_plot(self, mock_dumps, mock_get_template, mock_objects,
                  mock__process_tasks, **ddt_kwargs):
        mock__process_tasks.return_value = "source", "scenarios"
        mock_get_template.return_value.render.return_value = "tasks_html"
        mock_objects.Task.extend_results.return_value = ["extended_result"]
        tasks_results = [
            {"key": "foo_key", "sla": "foo_sla", "result": "foo_result",
             "full_duration": "foo_full_duration",
             "load_duration": "foo_load_duration"}]
        html = plot.plot(tasks_results, **ddt_kwargs)
        self.assertEqual(html, "tasks_html")
        generic_results = [
            {"id": None, "created_at": None, "updated_at": None,
             "task_uuid": None, "key": "foo_key",
             "data": {"raw": "foo_result",
                      "full_duration": "foo_full_duration",
                      "sla": "foo_sla",
                      "load_duration": "foo_load_duration"}}]
        mock_objects.Task.extend_results.assert_called_once_with(
            generic_results)
        mock_get_template.assert_called_once_with("task/report.html")
        mock__process_tasks.assert_called_once_with(["extended_result"])
        if "include_libs" in ddt_kwargs:
            mock_get_template.return_value.render.assert_called_once_with(
                data="json_scenarios", source="json_source",
                include_libs=ddt_kwargs["include_libs"])
        else:
            mock_get_template.return_value.render.assert_called_once_with(
                data="json_scenarios", source="json_source",
                include_libs=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock

from rally.plugins.openstack.scenarios.neutron import loadbalancer_v1
from tests.unit import test


@ddt.ddt
class NeutronLoadbalancerv1TestCase(test.TestCase):

    def _get_context(self):
        context = test.get_test_context()
        context.update({
            "user": {
                "id": "fake_user",
                "tenant_id": "fake_tenant",
                "credential": mock.MagicMock()
            },
            "tenant": {"id": "fake_tenant",
                       "networks": [{"id": "fake_net",
                                     "subnets": ["fake_subnet"]}]}})
        return context

    @ddt.data(
        {},
        {"pool_create_args": None},
        {"pool_create_args": {}},
        {"pool_create_args": {"name": "given-name"}},
    )
    @ddt.unpack
    def test_create_and_list_pools(self, pool_create_args=None):
        neutron_scenario = loadbalancer_v1.NeutronLoadbalancerV1(
            self._get_context())
        pool_data = pool_create_args or {}
        networks = self._get_context()["tenant"]["networks"]
        neutron_scenario._create_v1_pools = mock.Mock()
        neutron_scenario._list_v1_pools = mock.Mock()
        neutron_scenario.create_and_list_pools(
            pool_create_args=pool_create_args)
        neutron_scenario._create_v1_pools.assert_called_once_with(
            networks, **pool_data)
        neutron_scenario._list_v1_pools.assert_called_once_with()

    @ddt.data(
        {},
        {"pool_create_args": None},
        {"pool_create_args": {}},
        {"pool_create_args": {"name": "given-name"}},
    )
    @ddt.unpack
    def test_create_and_delete_pools(self, pool_create_args=None):
        neutron_scenario = loadbalancer_v1.NeutronLoadbalancerV1(
            self._get_context())
        pools = [{
            "pool": {
                "id": "pool-id"
            }
        }]
        pool_data = pool_create_args or {}
        networks = self._get_context()["tenant"]["networks"]
        neutron_scenario._create_v1_pools = mock.Mock(return_value=pools)
        neutron_scenario._delete_v1_pool = mock.Mock()
        neutron_scenario.create_and_delete_pools(
            pool_create_args=pool_create_args)
        self.assertEqual([mock.call(networks, **pool_data)],
                         neutron_scenario._create_v1_pools.mock_calls)
        for pool in pools:
            self.assertEqual(1, neutron_scenario._delete_v1_pool.call_count)

    @ddt.data(
        {},
        {"pool_create_args": None},
        {"pool_create_args": {}},
        {"pool_create_args": {"name": "given-name"}},
        {"pool_update_args": None},
        {"pool_update_args": {}},
        {"pool_update_args": {"name": "updated-name"}},
        {"pool_create_args": None, "pool_update_args": None},
        {"pool_create_args": {"name": "given-name"},
         "pool_update_args": {"name": "updated-name"}},
        {"pool_create_args": None,
         "pool_update_args": {"name": "updated-name"}},
        {"pool_create_args": None, "pool_update_args": {}},
        {"pool_create_args": {}, "pool_update_args": None},
    )
    @ddt.unpack
    def test_create_and_update_pools(self, pool_create_args=None,
                                     pool_update_args=None):
        neutron_scenario = loadbalancer_v1.NeutronLoadbalancerV1(
            self._get_context())
        pools = [{
            "pool": {
                "id": "pool-id"
            }
        }]
        updated_pool = {
            "pool": {
                "id": "pool-id",
                "name": "updated-pool",
                "admin_state_up": True
            }
        }
        pool_data = pool_create_args or {}
        pool_update_args = pool_update_args or {}
        pool_update_args.update({"name": "_updated", "admin_state_up": True})
        neutron_scenario._create_v1_pools = mock.Mock(return_value=pools)
        neutron_scenario._update_v1_pool = mock.Mock(
            return_value=updated_pool)
        networks = self._get_context()["tenant"]["networks"]
        neutron_scenario.create_and_update_pools(
            pool_create_args=pool_data,
            pool_update_args=pool_update_args)
        self.assertEqual([mock.call(networks, **pool_data)],
                         neutron_scenario._create_v1_pools.mock_calls)
        for pool in pools:
            neutron_scenario._update_v1_pool.assert_called_once_with(
                pool, **pool_update_args)

    @ddt.data(
        {},
        {"vip_create_args": None},
        {"vip_create_args": {}},
        {"vip_create_args": {"name": "given-vip-name"}},
        {"pool_create_args": None},
        {"pool_create_args": {}},
        {"pool_create_args": {"name": "given-pool-name"}},
    )
    @ddt.unpack
    def test_create_and_list_vips(self, pool_create_args=None,
                                  vip_create_args=None):
        neutron_scenario = loadbalancer_v1.NeutronLoadbalancerV1(
            self._get_context())
        pools = [{
            "pool": {
                "id": "pool-id"
            }
        }]
        vip_data = vip_create_args or {}
        pool_data = pool_create_args or {}
        networks = self._get_context()["tenant"]["networks"]
        neutron_scenario._create_v1_pools = mock.Mock(return_value=pools)
        neutron_scenario._create_v1_vip = mock.Mock()
        neutron_scenario._list_v1_vips = mock.Mock()
        neutron_scenario.create_and_list_vips(
            pool_create_args=pool_create_args, vip_create_args=vip_create_args)
        neutron_scenario._create_v1_pools.assert_called_once_with(
            networks, **pool_data)
        neutron_scenario._create_v1_vip.assert_has_calls(
            [mock.call(pool, **vip_data) for pool in pools])
        neutron_scenario._list_v1_vips.assert_called_once_with()

    @ddt.data(
        {},
        {"vip_create_args": None},
        {"vip_create_args": {}},
        {"vip_create_args": {"name": "given-name"}},
        {"pool_create_args": None},
        {"pool_create_args": {}},
        {"pool_create_args": {"name": "given-pool-name"}},
    )
    @ddt.unpack
    def test_create_and_delete_vips(self, pool_create_args=None,
                                    vip_create_args=None):
        pools = [{
            "pool": {
                "id": "pool-id"
            }
        }]
        vip = {
            "vip": {
                "id": "vip-id"
            }
        }
        neutron_scenario = loadbalancer_v1.NeutronLoadbalancerV1(
            self._get_context())
        vip_data = vip_create_args or {}
        pool_data = pool_create_args or {}
        networks = self._get_context()["tenant"]["networks"]
        neutron_scenario._create_v1_pools = mock.Mock(return_value=pools)
        neutron_scenario._create_v1_vip = mock.Mock(return_value=vip)
        neutron_scenario._delete_v1_vip = mock.Mock()
        neutron_scenario.create_and_delete_vips(
            pool_create_args=pool_create_args,
            vip_create_args=vip_create_args)
        neutron_scenario._create_v1_pools.assert_called_once_with(
            networks, **pool_data)
        neutron_scenario._create_v1_vip.assert_has_calls(
            [mock.call(pool, **vip_data) for pool in pools])
        neutron_scenario._delete_v1_vip.assert_has_calls(
            [mock.call(vip["vip"])])

    @ddt.data(
        {},
        {"vip_create_args": None},
        {"vip_create_args": {}},
        {"vip_create_args": {"name": "given-vip-name"}},
        {"pool_create_args": None},
        {"pool_create_args": {}},
        {"pool_create_args": {"name": "given-pool-name"}},
    )
    @ddt.unpack
    def test_create_and_update_vips(self, pool_create_args=None,
                                    vip_create_args=None,
                                    vip_update_args=None):
        neutron_scenario = loadbalancer_v1.NeutronLoadbalancerV1(
            self._get_context())
        pools = [{
            "pool": {
                "id": "pool-id",
            }
        }]
        expected_vip = {
            "vip": {
                "id": "vip-id",
                "name": "vip-name"
            }
        }
        updated_vip = {
            "vip": {
                "id": "vip-id",
                "name": "updated-vip-name"
            }
        }
        vips = [expected_vip]
        vip_data = vip_create_args or {}
        vip_update_data = vip_update_args or {}
        pool_data = pool_create_args or {}
        networks = self._get_context()["tenant"]["networks"]
        neutron_scenario._create_v1_pools = mock.Mock(return_value=pools)
        neutron_scenario._create_v1_vip = mock.Mock(return_value=expected_vip)
        neutron_scenario._update_v1_vip = mock.Mock(return_value=updated_vip)
        neutron_scenario.create_and_update_vips(
            pool_create_args=pool_create_args, vip_create_args=vip_create_args,
            vip_update_args=vip_update_args)
        neutron_scenario._create_v1_pools.assert_called_once_with(
            networks, **pool_data)
        neutron_scenario._create_v1_vip.assert_has_calls(
            [mock.call(pool, **vip_data) for pool in pools])
        neutron_scenario._update_v1_vip.assert_has_calls(
            [mock.call(vip, **vip_update_data) for vip in vips])

    @ddt.data(
        {},
        {"healthmonitor_create_args": None},
        {"healthmonitor_create_args": {}},
        {"healthmonitor_create_args": {"name": "given-name"}},
    )
    @ddt.unpack
    def test_create_and_list_healthmonitors(self,
                                            healthmonitor_create_args=None):
        neutron_scenario = loadbalancer_v1.NeutronLoadbalancerV1(
            self._get_context())
        hm_data = healthmonitor_create_args or {}
        neutron_scenario._create_v1_healthmonitor = mock.Mock()
        neutron_scenario._list_v1_healthmonitors = mock.Mock()
        neutron_scenario.create_and_list_healthmonitors(
            healthmonitor_create_args=healthmonitor_create_args)
        neutron_scenario._create_v1_healthmonitor.assert_called_once_with(
            **hm_data)
        neutron_scenario._list_v1_healthmonitors.assert_called_once_with()

    @ddt.data(
        {},
        {"healthmonitor_create_args": None},
        {"healthmonitor_create_args": {}},
        {"healthmonitor_create_args": {"name": "given-name"}},
    )
    @ddt.unpack
    def test_create_and_delete_healthmonitors(self,
                                              healthmonitor_create_args=None):
        neutron_scenario = loadbalancer_v1.NeutronLoadbalancerV1(
            self._get_context())
        hm = {"health_monitor": {"id": "hm-id"}}
        hm_data = healthmonitor_create_args or {}
        neutron_scenario._create_v1_healthmonitor = mock.Mock(return_value=hm)
        neutron_scenario._delete_v1_healthmonitor = mock.Mock()
        neutron_scenario.create_and_delete_healthmonitors(
            healthmonitor_create_args=healthmonitor_create_args)
        neutron_scenario._create_v1_healthmonitor.assert_called_once_with(
            **hm_data)
        neutron_scenario._delete_v1_healthmonitor.assert_called_once_with(
            neutron_scenario._create_v1_healthmonitor.return_value[
                "health_monitor"])

    @ddt.data(
        {},
        {"healthmonitor_create_args": None},
        {"healthmonitor_create_args": {}},
        {"healthmonitor_create_args": {"name": "given-name"}},
    )
    @ddt.unpack
    def test_create_and_update_healthmonitors(self,
                                              healthmonitor_create_args=None,
                                              healthmonitor_update_args=None):
        neutron_scenario = loadbalancer_v1.NeutronLoadbalancerV1(
            self._get_context())
        mock_random = loadbalancer_v1.random = mock.Mock()
        hm = {"healthmonitor": {"id": "hm-id"}}
        hm_data = healthmonitor_create_args or {}
        hm_update_data = healthmonitor_update_args or {
            "max_retries": mock_random.choice.return_value}
        neutron_scenario._create_v1_healthmonitor = mock.Mock(return_value=hm)
        neutron_scenario._update_v1_healthmonitor = mock.Mock()
        neutron_scenario.create_and_update_healthmonitors(
            healthmonitor_create_args=healthmonitor_create_args,
            healthmonitor_update_args=healthmonitor_update_args)
        neutron_scenario._create_v1_healthmonitor.assert_called_once_with(
            **hm_data)
        neutron_scenario._update_v1_healthmonitor.assert_called_once_with(
            neutron_scenario._create_v1_healthmonitor.return_value,
            **hm_update_data)



# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

from oslo_config import cfg

from rally.plugins.openstack import scenario
from rally.task import atomic
from rally.task import utils


EC2_BENCHMARK_OPTS = [
    cfg.FloatOpt(
        "ec2_server_boot_prepoll_delay",
        default=1.0,
        help="Time to sleep after boot before polling for status"
    ),
    cfg.FloatOpt(
        "ec2_server_boot_timeout",
        default=300.0,
        help="Server boot timeout"
    ),
    cfg.FloatOpt(
        "ec2_server_boot_poll_interval",
        default=1.0,
        help="Server boot poll interval"
    )
]

CONF = cfg.CONF
benchmark_group = cfg.OptGroup(name="benchmark",
                               title="benchmark options")
CONF.register_opts(EC2_BENCHMARK_OPTS, group=benchmark_group)


class EC2Scenario(scenario.OpenStackScenario):
    """Base class for EC2 scenarios with basic atomic actions."""

    @atomic.action_timer("ec2.list_servers")
    def _list_servers(self):
        """Returns user servers list."""
        return self.clients("ec2").get_only_instances()

    @atomic.action_timer("ec2.boot_servers")
    def _boot_servers(self, image_id, flavor_name,
                      instance_num=1, **kwargs):
        """Boot multiple servers.

        Returns when all the servers are actually booted and are in the
        "Running" state.

        :param image_id: ID of the image to be used for server creation
        :param flavor_name: Name of the flavor to be used for server creation
        :param instance_num: Number of instances to boot
        :param kwargs: Other optional parameters to boot servers

        :returns: List of created server objects
        """
        reservation = self.clients("ec2").run_instances(
            image_id=image_id,
            instance_type=flavor_name,
            min_count=instance_num,
            max_count=instance_num,
            **kwargs)
        servers = [instance for instance in reservation.instances]

        time.sleep(CONF.benchmark.ec2_server_boot_prepoll_delay)
        servers = [utils.wait_for(
            server,
            ready_statuses=["RUNNING"],
            update_resource=self._update_resource,
            timeout=CONF.benchmark.ec2_server_boot_timeout,
            check_interval=CONF.benchmark.ec2_server_boot_poll_interval
        ) for server in servers]
        return servers

    def _update_resource(self, resource):
        resource.update()
        return resource



# Copyright 2014: The Rally team
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Rally command: show """

from __future__ import print_function

from rally import api
from rally.cli import cliutils
from rally.cli import envutils
from rally.common.i18n import _
from rally.common import objects
from rally.common import utils
from rally import osclients


class ShowCommands(object):
    """[Deprecated since 0.2.0] Show resources.

    Set of commands that allow you to view resources, provided by OpenStack
    cloud represented by deployment.
    """

    def _print_header(self, resource_name, credentials):
        print(_("\n%(resource)s for user `%(user)s` in tenant `%(tenant)s`:")
              % {"resource": resource_name,
                 "user": credentials["username"],
                 "tenant": credentials["tenant_name"]})

    @staticmethod
    def _get_credentials(deployment):
        deployment = api.Deployment.get(deployment)
        # NOTE(andreykurilin): it is a bad practise to access to inner db_obj,
        # but we can do it here, since we are planning to deprecate and remove
        # this  command at all.
        admin = deployment.deployment.get("admin")
        credentials = [admin] if admin else []

        return credentials + deployment.deployment.get("users", [])

    @cliutils.args("--deployment", dest="deployment", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of a deployment.")
    @envutils.with_default_deployment(cli_arg_name="deployment")
    @cliutils.process_keystone_exc
    def images(self, deployment=None):
        """Display available images.

        :param deployment: UUID or name of a deployment
        """
        headers = ["UUID", "Name", "Size (B)"]
        mixed_case_fields = ["UUID", "Name"]
        float_cols = ["Size (B)"]
        formatters = dict(zip(float_cols,
                              [cliutils.pretty_float_formatter(col)
                               for col in float_cols]))

        for credential_dict in self._get_credentials(deployment):
            self._print_header("Images", credential_dict)
            table_rows = []

            clients = osclients.Clients(objects.Credential(**credential_dict))
            glance_client = clients.glance()
            for image in glance_client.images.list():
                data = [image.id, image.name, image.size]
                table_rows.append(utils.Struct(**dict(zip(headers, data))))

            cliutils.print_list(table_rows,
                                fields=headers,
                                formatters=formatters,
                                mixed_case_fields=mixed_case_fields)

    @cliutils.args("--deployment", dest="deployment", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of a deployment.")
    @envutils.with_default_deployment(cli_arg_name="deployment")
    @cliutils.process_keystone_exc
    def flavors(self, deployment=None):
        """Display available flavors.

        :param deployment: UUID or name of a deployment
        """
        headers = ["ID", "Name", "vCPUs", "RAM (MB)", "Swap (MB)", "Disk (GB)"]
        mixed_case_fields = ["ID", "Name", "vCPUs"]
        float_cols = ["RAM (MB)", "Swap (MB)", "Disk (GB)"]
        formatters = dict(zip(float_cols,
                              [cliutils.pretty_float_formatter(col)
                               for col in float_cols]))

        for credential_dict in self._get_credentials(deployment):
            self._print_header("Flavors", credential_dict)
            table_rows = []
            clients = osclients.Clients(objects.Credential(**credential_dict))
            nova_client = clients.nova()
            for flavor in nova_client.flavors.list():
                data = [flavor.id, flavor.name, flavor.vcpus,
                        flavor.ram, flavor.swap, flavor.disk]
                table_rows.append(utils.Struct(**dict(zip(headers, data))))

            cliutils.print_list(table_rows,
                                fields=headers,
                                formatters=formatters,
                                mixed_case_fields=mixed_case_fields)

    @cliutils.args("--deployment", dest="deployment", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of a deployment.")
    @envutils.with_default_deployment(cli_arg_name="deployment")
    @cliutils.process_keystone_exc
    def networks(self, deployment=None):
        """Display configured networks."""

        headers = ["ID", "Label", "CIDR"]
        mixed_case_fields = ["ID", "Label", "CIDR"]

        for credential_dict in self._get_credentials(deployment):
            self._print_header("Networks", credential_dict)
            table_rows = []
            clients = osclients.Clients(objects.Credential(**credential_dict))
            nova_client = clients.nova()
            for network in nova_client.networks.list():
                data = [network.id, network.label, network.cidr]
                table_rows.append(utils.Struct(**dict(zip(headers, data))))

            cliutils.print_list(table_rows,
                                fields=headers,
                                mixed_case_fields=mixed_case_fields)

    @cliutils.args("--deployment", dest="deployment", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of a deployment.")
    @envutils.with_default_deployment(cli_arg_name="deployment")
    @cliutils.process_keystone_exc
    def secgroups(self, deployment=None):
        """Display security groups."""

        headers = ["ID", "Name", "Description"]
        mixed_case_fields = ["ID", "Name", "Description"]
        for credential_dict in self._get_credentials(deployment):
            self._print_header("Security groups", credential_dict)
            table_rows = []
            clients = osclients.Clients(objects.Credential(**credential_dict))
            nova_client = clients.nova()
            for secgroup in nova_client.security_groups.list():
                data = [secgroup.id, secgroup.name,
                        secgroup.description]
                table_rows.append(utils.Struct(**dict(zip(headers,
                                                          data))))
            cliutils.print_list(
                table_rows,
                fields=headers,
                mixed_case_fields=mixed_case_fields)

    @cliutils.args("--deployment", dest="deployment", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of a deployment.")
    @envutils.with_default_deployment(cli_arg_name="deployment")
    @cliutils.process_keystone_exc
    def keypairs(self, deployment=None):
        """Display available ssh keypairs."""

        headers = ["Name", "Fingerprint"]
        mixed_case_fields = ["Name", "Fingerprint"]

        for credential_dict in self._get_credentials(deployment):
            self._print_header("Keypairs", credential_dict)
            table_rows = []
            clients = osclients.Clients(objects.Credential(**credential_dict))
            nova_client = clients.nova()
            for keypair in nova_client.keypairs.list():
                data = [keypair.name, keypair.fingerprint]
                table_rows.append(utils.Struct(**dict(zip(headers, data))))
            cliutils.print_list(table_rows,
                                fields=headers,
                                mixed_case_fields=mixed_case_fields)



# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import copy

import mock

from rally.plugins.openstack.context.nova import servers
from tests.unit import fakes
from tests.unit import test

CTX = "rally.plugins.openstack.context.nova"
SCN = "rally.plugins.openstack.scenarios"
TYP = "rally.task.types"


class ServerGeneratorTestCase(test.ScenarioTestCase):

    def _gen_tenants(self, count):
        tenants = {}
        for id_ in range(count):
            tenants[str(id_)] = {"name": str(id_)}
        return tenants

    def test_init(self):
        tenants_count = 2
        servers_per_tenant = 5
        self.context.update({
            "config": {
                "servers": {
                    "servers_per_tenant": servers_per_tenant,
                }
            },
            "tenants": self._gen_tenants(tenants_count)})

        inst = servers.ServerGenerator(self.context)
        self.assertEqual(self.context["config"]["servers"], inst.config)

    @mock.patch("%s.nova.utils.NovaScenario._boot_servers" % SCN,
                return_value=[
                    fakes.FakeServer(id="uuid"),
                    fakes.FakeServer(id="uuid"),
                    fakes.FakeServer(id="uuid"),
                    fakes.FakeServer(id="uuid"),
                    fakes.FakeServer(id="uuid")
                ])
    @mock.patch("%s.ImageResourceType.transform" % TYP,
                return_value=mock.MagicMock())
    @mock.patch("%s.FlavorResourceType.transform" % TYP,
                return_value=mock.MagicMock())
    @mock.patch("%s.servers.osclients" % CTX, return_value=fakes.FakeClients())
    def test_setup(self, mock_osclients, mock_flavor_resource_type_transform,
                   mock_image_resource_type_transform,
                   mock_nova_scenario__boot_servers):

        tenants_count = 2
        users_per_tenant = 5
        servers_per_tenant = 5

        tenants = self._gen_tenants(tenants_count)
        users = []
        for id_ in tenants.keys():
            for i in range(users_per_tenant):
                users.append({"id": i, "tenant_id": id_,
                              "credential": mock.MagicMock()})

        self.context.update({
            "config": {
                "users": {
                    "tenants": 2,
                    "users_per_tenant": 5,
                    "concurrent": 10,
                },
                "servers": {
                    "auto_assign_nic": True,
                    "servers_per_tenant": 5,
                    "image": {
                        "name": "cirros-0.3.4-x86_64-uec",
                    },
                    "flavor": {
                        "name": "m1.tiny",
                    },
                },
            },
            "admin": {
                "credential": mock.MagicMock()
            },
            "users": users,
            "tenants": tenants
        })

        new_context = copy.deepcopy(self.context)
        for id_ in new_context["tenants"]:
            new_context["tenants"][id_].setdefault("servers", [])
            for i in range(servers_per_tenant):
                new_context["tenants"][id_]["servers"].append("uuid")

        servers_ctx = servers.ServerGenerator(self.context)
        servers_ctx.setup()
        self.assertEqual(new_context, self.context)
        image_id = mock_image_resource_type_transform.return_value
        flavor_id = mock_flavor_resource_type_transform.return_value
        servers_ctx_config = self.context["config"]["servers"]
        expected_auto_nic = servers_ctx_config.get("auto_assign_nic", False)
        expected_requests = servers_ctx_config.get("servers_per_tenant", False)
        called_times = len(tenants)
        mock_calls = [mock.call(image_id, flavor_id,
                                auto_assign_nic=expected_auto_nic,
                                requests=expected_requests)
                      for i in range(called_times)]
        mock_nova_scenario__boot_servers.assert_has_calls(mock_calls)

    @mock.patch("%s.servers.osclients" % CTX)
    @mock.patch("%s.servers.resource_manager.cleanup" % CTX)
    def test_cleanup(self, mock_cleanup, mock_osclients):

        tenants_count = 2
        users_per_tenant = 5
        servers_per_tenant = 5

        tenants = self._gen_tenants(tenants_count)
        users = []
        for id_ in tenants.keys():
            for i in range(users_per_tenant):
                users.append({"id": i, "tenant_id": id_,
                              "credential": "credential"})
            tenants[id_].setdefault("servers", [])
            for j in range(servers_per_tenant):
                tenants[id_]["servers"].append("uuid")

        self.context.update({
            "config": {
                "users": {
                    "tenants": 2,
                    "users_per_tenant": 5,
                    "concurrent": 10,
                },
                "servers": {
                    "servers_per_tenant": 5,
                    "image": {
                        "name": "cirros-0.3.4-x86_64-uec",
                    },
                    "flavor": {
                        "name": "m1.tiny",
                    },
                },
            },
            "admin": {
                "credential": mock.MagicMock()
            },
            "users": users,
            "tenants": tenants
        })

        servers_ctx = servers.ServerGenerator(self.context)
        servers_ctx.cleanup()

        mock_cleanup.assert_called_once_with(names=["nova.servers"],
                                             users=self.context["users"])



# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from rally.common import logging
from rally import consts
from rally.plugins.openstack import scenario
from rally.plugins.openstack.scenarios.ec2 import utils
from rally.task import types
from rally.task import validation


LOG = logging.getLogger(__name__)


class EC2Servers(utils.EC2Scenario):
    """Benchmark scenarios for servers using EC2."""

    @validation.required_services(consts.Service.EC2)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"cleanup": ["ec2"]})
    def list_servers(self):
        """List all servers.

        This simple scenario tests the EC2 API list function by listing
        all the servers.
        """
        self._list_servers()

    @types.set(image=types.EC2ImageResourceType,
               flavor=types.EC2FlavorResourceType)
    @validation.image_valid_on_flavor("flavor", "image")
    @validation.required_services(consts.Service.EC2)
    @validation.required_openstack(users=True)
    @scenario.configure(context={"cleanup": ["ec2"]})
    def boot_server(self, image, flavor, **kwargs):
        """Boot a server.

        Assumes that cleanup is done elsewhere.

        :param image: image to be used to boot an instance
        :param flavor: flavor to be used to boot an instance
        :param kwargs: optional additional arguments for server creation
        """
        self._boot_servers(image, flavor, **kwargs)



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import math

import six

from rally.task.processing import utils


@six.add_metaclass(abc.ABCMeta)
class StreamingAlgorithm(object):
    """Base class for streaming computations that scale."""

    @abc.abstractmethod
    def add(self, value):
        """Process a single value from the input stream."""

    @abc.abstractmethod
    def merge(self, other):
        """Merge results processed by another instance."""

    @abc.abstractmethod
    def result(self):
        """Return the result based on the values processed so far."""

    def _cast_to_float(self, value):
        try:
            return float(value)
        except (TypeError, ValueError):
            raise TypeError("Non-numerical value: %r" % value)


class MeanComputation(StreamingAlgorithm):
    """Compute mean for a stream of numbers."""

    def __init__(self):
        self.total = 0.0
        self.count = 0

    def add(self, value):
        self.count += 1
        self.total += value

    def merge(self, other):
        self.count += other.count
        self.total += other.total

    def result(self):
        if self.count:
            return self.total / self.count
        return None


class StdDevComputation(StreamingAlgorithm):
    """Compute standard deviation for a stream of numbers."""

    def __init__(self):
        self.count = 0
        # NOTE(msdubov): To compute std, we need the auxiliary variables below.
        self.dev_sum = 0.0
        self.mean_computation = MeanComputation()
        self.mean = 0.0

    def add(self, value):
        # NOTE(msdubov): This streaming method for std computation appears
        #                in "The Art of Computer Programming" by D. Knuth,
        #                Vol 2, p. 232, 3rd edition.
        self.count += 1
        mean_prev = self.mean
        self.mean_computation.add(value)
        self.mean = self.mean_computation.result()
        self.dev_sum = self.dev_sum + (value - mean_prev) * (value - self.mean)

    def merge(self, other):
        if not other.mean_computation.count:
            return
        dev_sum1 = self.dev_sum
        count1 = self.count
        mean1 = self.mean

        dev_sum2 = other.dev_sum
        count2 = other.count
        mean2 = other.mean

        self.mean_computation.merge(other.mean_computation)
        self.mean = self.mean_computation.result()
        self.count += other.count

        self.dev_sum = (dev_sum1 + count1 * mean1 ** 2 +
                        dev_sum2 + count2 * mean2 ** 2 -
                        self.count * self.mean ** 2)

    def result(self):
        # NOTE(amaretskiy): Need at least two values to be processed
        if self.count < 2:
            return None
        return math.sqrt(self.dev_sum / (self.count - 1))


class MinComputation(StreamingAlgorithm):
    """Compute minimal value from a stream of numbers."""

    def __init__(self):
        self._value = None

    def add(self, value):
        value = self._cast_to_float(value)

        if self._value is None or value < self._value:
            self._value = value

    def merge(self, other):
        if other._value is not None:
            self.add(other._value)

    def result(self):
        return self._value


class MaxComputation(StreamingAlgorithm):
    """Compute maximal value from a stream of numbers."""

    def __init__(self):
        self._value = None

    def add(self, value):
        value = self._cast_to_float(value)

        if self._value is None or value > self._value:
            self._value = value

    def merge(self, other):
        if other._value is not None:
            self.add(other._value)

    def result(self):
        return self._value


class PercentileComputation(StreamingAlgorithm):
    """Compute percentile value from a stream of numbers."""

    def __init__(self, percent, length):
        """Init streaming computation.

        :param percent: numeric percent (from 0.00..1 to 0.999..)
        :param length: count of the measurements
        """
        if not 0 < percent < 1:
            raise ValueError("Unexpected percent: %s" % percent)
        self._percent = percent

        self._graph_zipper = utils.GraphZipper(length, 10000)

    def add(self, value):
        self._graph_zipper.add_point(value)

    def merge(self, other):
        # TODO(ikhudoshyn): Implement me
        raise NotImplementedError()

    def result(self):
        results = list(
            map(lambda x: x[1], self._graph_zipper.get_zipped_graph()))
        if results:
            return utils.percentile(results, self._percent)
        return None


class IncrementComputation(StreamingAlgorithm):
    """Simple incremental counter."""

    def __init__(self):
        self._count = 0

    def add(self, *args):
        self._count += 1

    def merge(self, other):
        self._count += other._count

    def result(self):
        return self._count



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally.common.i18n import _
from rally.common import logging
from rally import consts
from rally import exceptions
from rally import osclients
from rally.plugins.openstack.wrappers import keystone
from rally.task import context

LOG = logging.getLogger(__name__)


@context.configure(name="roles", order=330)
class RoleGenerator(context.Context):
    """Context class for adding temporary roles for benchmarks."""

    CONFIG_SCHEMA = {
        "type": "array",
        "$schema": consts.JSON_SCHEMA,
        "items": {
            "type": "string",
        },
        "additionalProperties": False
    }

    def __init__(self, ctx):
        super(RoleGenerator, self).__init__(ctx)
        self.credential = self.context["admin"]["credential"]

    def _add_role(self, admin_credential, context_role):
        """Add role to users.

        :param admin_credential: The admin credential.
        :param context_role: name of existing role.
        """
        client = keystone.wrap(osclients.Clients(admin_credential).keystone())
        default_roles = client.list_roles()
        for def_role in default_roles:
            if str(def_role.name) == context_role:
                role = def_role
                break
        else:
            raise exceptions.NoSuchRole(role=context_role)

        LOG.debug("Adding role %s to all users" % (role.id))
        for user in self.context["users"]:
            client.add_role(user_id=user["id"], role_id=role.id,
                            project_id=user["tenant_id"])

        return {"id": str(role.id), "name": str(role.name)}

    def _remove_role(self, admin_credential, role):
        """Remove given role from users.

        :param admin_credential: The admin credential.
        :param role: dictionary with role parameters (id, name).
        """
        client = keystone.wrap(osclients.Clients(admin_credential).keystone())

        for user in self.context["users"]:
            with logging.ExceptionLogger(
                    LOG, _("Failed to remove role: %s") % role["id"]):
                client.remove_role(
                    user_id=user["id"], role_id=role["id"],
                    project_id=user["tenant_id"])

    @logging.log_task_wrapper(LOG.info, _("Enter context: `roles`"))
    def setup(self):
        """Add roles to all users."""
        self.context["roles"] = [self._add_role(self.credential, name)
                                 for name in self.config]

    @logging.log_task_wrapper(LOG.info, _("Exit context: `roles`"))
    def cleanup(self):
        """Remove roles from users."""
        for role in self.context["roles"]:
            self._remove_role(self.credential, role)



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from rally.plugins.openstack.scenarios.tempest import tempest
from rally.verification.tempest import tempest as verifier
from tests.unit import test

TEMPEST_DIR = "rally.verification.tempest"
VERIFIER = TEMPEST_DIR + ".tempest"
TS = "rally.plugins.openstack.scenarios.tempest"


class TempestScenarioTestCase(test.TestCase):

    def setUp(self):
        super(TempestScenarioTestCase, self).setUp()
        self.verifier = verifier.Tempest("fake_uuid")
        self.verifier.log_file_raw = "/dev/null"
        self.verifier.parse_results = mock.MagicMock()
        self.verifier.parse_results.return_value = mock.MagicMock(
            tests={}, total={"time": 0})
        self.context = test.get_test_context()
        self.context.update({"verifier": self.verifier,
                             "tmp_results_dir": "/dev"})
        self.scenario = tempest.TempestScenario(self.context)
        self.scenario._add_atomic_actions = mock.MagicMock()

    def get_tests_launcher_cmd(self, tests):
        return ("%(venv)s testr run --subunit --parallel %(tests)s "
                "| tee /dev/null "
                "| %(venv)s subunit-trace -f -n" %
                {
                    "venv": self.verifier.venv_wrapper,
                    "tests": " ".join(tests)
                })

    @mock.patch(TS + ".utils.tempfile")
    @mock.patch(VERIFIER + ".subprocess")
    @mock.patch(TEMPEST_DIR + ".config.TempestResourcesContext")
    def test_single_test(self, mock_tempest_resources_context,
                         mock_subprocess, mock_tempfile):
        mock_tempfile.NamedTemporaryFile().name = "/dev/null"
        fake_test = "tempest.api.fake.test"

        self.scenario.single_test(test_name=fake_test)

        expected_call = self.get_tests_launcher_cmd([fake_test])
        mock_subprocess.check_call.assert_called_once_with(
            expected_call, cwd=self.verifier.path(),
            env=self.verifier.env, shell=True)

    @mock.patch(TS + ".utils.tempfile")
    @mock.patch(VERIFIER + ".subprocess")
    @mock.patch(TEMPEST_DIR + ".config.TempestResourcesContext")
    def test_single_test_negative(self, mock_tempest_resources_context,
                                  mock_subprocess, mock_tempfile):
        mock_tempfile.NamedTemporaryFile().name = "/dev/null"
        fake_test = "tempest.api.network"

        self.scenario.single_test(test_name=fake_test)

        expected_call = self.get_tests_launcher_cmd([fake_test])
        mock_subprocess.check_call.assert_called_once_with(
            expected_call, cwd=self.verifier.path(),
            env=self.verifier.env, shell=True)

    @mock.patch(TS + ".utils.tempfile")
    @mock.patch(VERIFIER + ".subprocess")
    @mock.patch(TEMPEST_DIR + ".config.TempestResourcesContext")
    def test_single_test_without_prefix(self, mock_tempest_resources_context,
                                        mock_subprocess, mock_tempfile):
        mock_tempfile.NamedTemporaryFile().name = "/dev/null"

        self.scenario.single_test("network")

        expected_call = self.get_tests_launcher_cmd(["tempest.api.network"])
        mock_subprocess.check_call.assert_called_once_with(
            expected_call, cwd=self.verifier.path(),
            env=self.verifier.env, shell=True)

    @mock.patch(TS + ".utils.tempfile")
    @mock.patch(VERIFIER + ".subprocess")
    @mock.patch(TEMPEST_DIR + ".config.TempestResourcesContext")
    def test_all(self, mock_tempest_resources_context,
                 mock_subprocess, mock_tempfile):
        mock_tempfile.NamedTemporaryFile().name = "/dev/null"

        self.scenario.all()

        expected_call = self.get_tests_launcher_cmd([])
        mock_subprocess.check_call.assert_called_once_with(
            expected_call, cwd=self.verifier.path(),
            env=self.verifier.env, shell=True)

    @mock.patch(TS + ".utils.tempfile")
    @mock.patch(VERIFIER + ".subprocess")
    @mock.patch(TEMPEST_DIR + ".config.TempestResourcesContext")
    def test_set_smoke(self, mock_tempest_resources_context,
                       mock_subprocess, mock_tempfile):
        mock_tempfile.NamedTemporaryFile().name = "/dev/null"

        self.scenario.set("smoke")

        expected_call = self.get_tests_launcher_cmd(["smoke"])
        mock_subprocess.check_call.assert_called_once_with(
            expected_call, cwd=self.verifier.path(),
            env=self.verifier.env, shell=True)

    @mock.patch(TS + ".utils.tempfile")
    @mock.patch(VERIFIER + ".subprocess")
    @mock.patch(TEMPEST_DIR + ".config.TempestResourcesContext")
    def test_set_full(self, mock_tempest_resources_context,
                      mock_subprocess, mock_tempfile):
        mock_tempfile.NamedTemporaryFile().name = "/dev/null"

        self.scenario.set("full")

        expected_call = self.get_tests_launcher_cmd([])
        mock_subprocess.check_call.assert_called_once_with(
            expected_call, cwd=self.verifier.path(),
            env=self.verifier.env, shell=True)

    @mock.patch(TS + ".utils.tempfile")
    @mock.patch(TEMPEST_DIR + ".config.TempestResourcesContext")
    def test_set_from_list(self, mock_tempest_resources_context,
                           mock_tempfile):
        mock_tempfile.NamedTemporaryFile().name = "/dev/null"

        fake_scenarios = ["network", "volume", "baremetal",
                          "image", "identity", "compute", "database",
                          "data_processing", "object_storage",
                          "telemetry", "queuing", "orchestration"]
        for fake_scenario in fake_scenarios:
            with mock.patch(VERIFIER + ".subprocess") as mock_subprocess:
                self.scenario.set(fake_scenario)
                fake_test = "tempest.api." + fake_scenario

                expected_call = self.get_tests_launcher_cmd([fake_test])
                mock_subprocess.check_call.assert_called_once_with(
                    expected_call, cwd=self.verifier.path(),
                    env=self.verifier.env, shell=True)

    @mock.patch(TS + ".utils.tempfile")
    @mock.patch(VERIFIER + ".subprocess")
    @mock.patch(TEMPEST_DIR + ".config.TempestResourcesContext")
    def test_set_selective(self, mock_tempest_resources_context,
                           mock_subprocess, mock_tempfile):
        mock_tempfile.NamedTemporaryFile().name = "/dev/null"

        self.scenario.set("network")

        expected_call = self.get_tests_launcher_cmd(["tempest.api.network"])
        mock_subprocess.check_call.assert_called_once_with(
            expected_call, cwd=self.verifier.path(),
            env=self.verifier.env, shell=True)

    @mock.patch(TS + ".utils.tempfile")
    @mock.patch(VERIFIER + ".subprocess")
    @mock.patch(TEMPEST_DIR + ".config.TempestResourcesContext")
    def test_list_of_tests(self, mock_tempest_resources_context,
                           mock_subprocess, mock_tempfile):
        mock_tempfile.NamedTemporaryFile().name = "/dev/null"
        fake_tests = ["tempest.fake.test1", "tempest.fake.test2"]

        self.scenario.list_of_tests(fake_tests)

        expected_call = self.get_tests_launcher_cmd(fake_tests)
        mock_subprocess.check_call.assert_called_once_with(
            expected_call, cwd=self.verifier.path(),
            env=self.verifier.env, shell=True)

    @mock.patch(TS + ".utils.tempfile")
    @mock.patch(VERIFIER + ".subprocess")
    @mock.patch(TEMPEST_DIR + ".config.TempestResourcesContext")
    def test_specific_regex(self, mock_tempest_resources_context,
                            mock_subprocess, mock_tempfile):
        mock_tempfile.NamedTemporaryFile().name = "/dev/null"
        regex = "tempest.fake.test1"

        self.scenario.specific_regex(regex)

        expected_call = self.get_tests_launcher_cmd([regex])
        mock_subprocess.check_call.assert_called_once_with(
            expected_call, cwd=self.verifier.path(),
            env=self.verifier.env, shell=True)
